define(function (require) {

    var Instance = require("entities/Instance");
    var Deferred = require("js/Deferred");
    
    var ajax = require("jquery").ajax;
    var ROOT = "/office-rest/rest/v1/persistence";

	function getQueryData(path, attributes, criteria) {
		/*- Copy the criteria object */
    	var r = {};//js.mixIn(criteria);
    	
    	/*- Allow usage like: Query.execute({attributes: "type,count:id count"}) */
    	if(typeof attributes === "string") {
    		attributes = attributes.split(",");
    	} 
    	
		r.fields = {};
		(attributes || []).forEach(function(a) {
			a = a.split(" ");
			r.fields[a[a.length === 2 ? 1 : 0]] = a.shift();
		});
    	
    	if(criteria && criteria.where) {
    		r.filterExpression = criteria.where;
    	}
    	
    	if(criteria && criteria.having) {
    		r.having = criteria.having;
    	}
    	
    	/*- Convert string usage */
    	if(criteria && typeof criteria.groupBy === "string") {
    		r.groupBy = criteria.groupBy.split(",");
    	}
    	
    	/* Convert string usage */
    	if(criteria && typeof criteria.orderBy === "string") {
    		r.orderBy = criteria.orderBy.split(",").map(function(ob) {
    			ob = ob.split(" ");
    			return { path: ob[0], type: ob[1] || "asc" };
    		});
    	}
    	
    	if(criteria && criteria.hasOwnProperty("page")) {
    		if((r.page = criteria.page) instanceof Array) {
	    		r.pagesize = r.page[1] || 50;
	    		r.page = r.page[0];
    		} else {
    			r.pagesize = 50;
    		}
    	} else {
			r.pagesize = 50;
    	}
    	
    	if(criteria && criteria.hasOwnProperty("pagesize")) {
    		r.pagesize = criteria.pagesize;
    	}
    	
    	return JSON.stringify(r);
	}
   
    return {
    	instances: {},
    	
    	getFotos: function(path) {
    		var d = new Deferred();
    		ajax({ url: ROOT + "persistence/" + path })
    			.then(function(res) { d.callback(JSON.parse(res)); })
    			.fail(function(err) { 
    				console.log(">>> persistence/" + path, err);
    				d.errback(err); 
    			});
    		return d;
    	},
    	update: function(path, key, values) {
    		var d = new Deferred();
    		ajax({
    			url: String.format("%s/%s/%s", ROOT, path, key), 
    			method: "POST",
		        contentType: "application/json",
		        data: JSON.stringify(values),
		        success: function (res) { d.callback(res); },
		        error: function (res) { d.errback(res); }
    		});
    		return d;
    	},
    	count: function(path, criteria) {
    		var attrs = (criteria||{}).attributes || "count:id";
    		return this.query(path, attrs, criteria)
    			.addCallback(function(res) {
    				return res[0][attrs.split(" ").pop()];
    			});
    	},
        query: function(path, attributes, criteria) {
        	var d = new Deferred(), EM = this;
        	
			criteria = js.mixIn(criteria);
			if(!criteria.pagesize) {
				criteria.pagesize = 50;
			}
			
        	if(criteria.count) {
        		var count_criteria = criteria.count === true ? 
        			criteria : criteria.count;
				delete criteria.count;
				
    			EM.count(path, count_criteria)
    				.addCallback(function(res) {
	    				var count = res;
	    				EM.query(path, attributes, criteria)
		    				.addCallback(function(res) {
		    					d.callback([count, res]);
		    				});
	    			})
	    			.addErrback(function(err) {
	    				d.errback(err);
	    			});

        	} else {
			    ajax({
			        url: String.format("%s/%s/query", ROOT, path),
			        method: "POST",
			        contentType: "application/json",
			        data: getQueryData(path, attributes, criteria),
			        success: function (res) {
	/*
			        	if(attributes.indexOf(".") !== -1) {
			        		attributes = attributes.split(",").filter(function(s) {
			        			return s.indexOf(" ") === -1 && s.indexOf(".") !== -1;
			        		});
			        		attributes.length && res.forEach(function(tuple, i) {
				        		attributes.forEach(function(a) {
				        			js.set(a, tuple[a], tuple);
				        			delete tuple[a];
				        		});
				        	});
			        	}
	*/
			            d.callback(res);
			        },
			        error: function (res) {
			        	var e = new Error(res.status);
			        	e.cause = res.responseJSON || res.responseText;
			            d.errback(e);
			        }
			    });
        	}
		    return d;
		},
		get: function(path) {
			var Session = require("veldoffice/Session");
			return Session.request("persistence/" + path);
		},
		fetch: function(path) {
			
		},
		
    	getFotos_p: function(path) {
    		return new Promise(function(resolve, reject) {
	    		ajax({ url: ROOT + "persistence/" + path })
	    			.then(
	    				(res) => resolve(JSON.parse(res)), 
	    				reject
	    			);
    		});
    	},
    	update_p: function(path, key, values) {
    		return new Promise(function(resolve, reject) {
	    		ajax({
	    			url: String.format("%s/%s/%s", ROOT, path, key), 
	    			method: "POST",
			        contentType: "application/json",
			        data: JSON.stringify(values),
			        success: resolve,
			        error: reject
	    		});
    		});
    	},
		query_p: function(path, attributes, criteria) {
        	var EM = this; criteria = js.mixIn(criteria);
    		var count_criteria = criteria.count === true ? criteria : criteria.count;
			delete criteria.count;
        	
        	return new Promise((resolve, reject) => {
	        	if(count_criteria) {
	    			EM.count_p(path, count_criteria)
	    				.then(
	    					(count) => EM.query_p(path, attributes, criteria)
			    				.then(
			    					(objs) => resolve([count, objs]), 
			    					(e) => reject(e)
			    				),
			    			(e) => reject(e)
			    		);
	        	} else {
				    ajax({
				        url: String.format("%s/%s/query", ROOT, path),
				        method: "POST",
				        contentType: "application/json",
				        data: getQueryData(path, attributes, criteria)
				    }).then(resolve).catch(reject);
	        	}
        	});
		},
    	count_p: function(path, criteria) {
    		var attrs = (criteria||{}).attributes || "count:id";
    		return this.query_p(path, attrs, criteria)
    			.then(function(res) {
    				return res[0][attrs.split(" ").pop()];
    			});
    	},
    	get_p: function(path) {
        	return new Promise((resolve, reject) => {
        		ajax({
        			url: ROOT + path,
        			method: "GET",
			        // contentType: "application/json",
			        // data: getQueryData(path, attributes, criteria),
			        success: resolve,
			        error: reject
        		});
        	});	
    	},

		reduce_or: function(acc, current, index, array) {
			return index ? ["or", current, acc] : current;
		},
		
		processWalkResult: function(result) {
			result.objs.forEach(obj => {
				for(var k in obj) {
					if(obj[k] instanceof Array) {
						if(obj[k].length > 0) {
							var elem0 = obj[k][0];
							if(elem0 instanceof Array) {
								// resolve an array of references
								obj[k] = obj[k].map(_ => result.objs[_]);
							} else {
								// resolve single reference (TODO maybe attribute is Array-like)
								obj[k] = result.objs[elem0];
							}
						} else {
							// empty set, no change necessary
						}
					}
				}
			});
			var instances = this.instances;
			for(var entity in result.instances) {
				if(instances[entity] === undefined) {
					instances[entity] = {};
				}
				for(var key in result.instances[entity]) {
					var index = result.instances[entity][key];
					instances[entity][key] = result.objs[index];
					// instances[entity][key]['@key'] = key;
				}
				
			}
			return result;
		},
		processQueryResult: function(entity, result) {
			
		}
    };

});

	/*-
		Hey man,
		
		De applicatie is nu in 2'en gedeeld:
		1) de back-end (het oude veldoffice project, development deploy versie meegeleverd)
		2) de front-end (source-versie): veldoffice2-front-end, zie svn
		
		In principe is de front-end qua inhoud hetzelfde, alleen bevat het nu bower en grunt configuraties. De hele bende wordt samengepakt etc. maar volgens mij ben je er wel mee bekend, dus als je de gruntfile.js en bower.json bekijkt zullen de details wel duidelijk worden. Het oude office-front-end project wordt vanaf nu niet meer gebruikt.
		
		Voor de profielen: omdat het aanzwengelen van het script in de productieversie niet zou werken (harde referentie naar c:/data/ ...) heb ik er een setting van gemaakt. In de back-end war zit een development.settings.properties (veldoffice-ejb/config) die je kunt aanpassen om zo je lokale script aan te roepen:
		settings.development.properties -> profielen.script=c:/data/office/profielen/run.js
		
		Op de back-end is nu ook een query interface beschikbaar: 
		POST: /<entity>/query of /<entity1>/<entity2>/.../<entityn>/query
		
		Post een object met het volgende formaat:
		- fields: array van de te selecteren velden: alias, path, bijv. { "pc": "projectcode", "mp": "countDistinct:meetpunten"}
		- filters: array van filter criteria, { path, type, value }: { path: "bodemlagen.onderkant", type: "gte", value: 10}
		- orderBy: array van sorteervolgordes, { path, type (desc, asc)} bijv. {path: "meetpunten.created", type: "desc"}, asc is default
		- groupBy: array van group by, bijv: meetpunten.type
		- having: array van expressies voor having van de vorm: <path><operator><value>, bijv. countDistinct:id>3
		- pagesize: paginagrootte
		- page: de op te vragen pagina
		
		filters types: contains, endswith, equals, gte, inlist, isempty, isnotnull, isnull, lt, lte, notequals, startswith
		aggregates: <func>:<path>, bijv (/onderzoek/query): countDistinct:meetpunten, max:meetpunten.bodemlagen.bovenkant
		
		Voor alle paden gelden de namen zoals gedefinieerd in het entity model. Onderzoek -> meetpunt -> bodemlaag is: 'meetpunten.bodemlagen'.
		
		Als je geen admin bent, wordt er automatisch gefilterd op bedrijf voor zover er een hierarchische relatie met de gevraagde entiteit bestaat.
		
		Simpel voorbeeldje:
		/onderzoek/query
		{
			fields: {
				"pc": "projectcode",
				"naam": "naam",
				"bodemlaag": "max:meetpunten.bodemlagen.onderkant"
			}, 
			groupBy: ["id"],
			orderBy: [{
					path: "naam",
					type: "desc"
			}],
			pagesize: 25,
			page: 3
		}
		
		Zie de front-end /test voor een kleine demo.
		
		Het is een eerste versie, dus het zou me niets verbazen wanneer er hier en daar nog wat bugs/fouten inzitten. Laat me weten hoe 't gaat.
		
		mzzls,
		Arnoud
	*/
	/*
		Standaard acties voor elke entiteit

		GET: /entity -> search
		
		optional parameters:
		pagesize: pagesize
		page: op te vragen pagina
		e: expansion 
		v: view -> voorgedefinieerd view (let op: views ondersteunen geen expansions)
		path=value -> selectiecriteria, voorbeelden: naam=vijfhuizen, tov.id=1234, ...
		sortby: pad om te sorteren, default = id
		sorttype: ASC, DESC, default = ASC
		
		GET: /entity/count
		Geeft het aantal terug, accepteert dezelfde filters als search
		
		
		GET: /entity/group
		overzicht van alle ManyToOne referenties van de selectie met de aantalen waarin ze voorkomen (gebruikt voor mogelijke filterwaarden in de grids)
		bijv. /onderzoek/2951/meetpunt/group
		[
		  {
		    "name": "tov",
		    "filters": [
		      {
		        "name": "NAP",
		        "count": 33,
		        "id": 2512
		      }
		    ]
		  },
		  {
		    "name": "type",
		    "filters": [
		      {
		        "name": "grondboring",
		        "count": 33,
		        "id": 2377
		      }
		    ]
		  },
		  {
		    "name": "situatiebeschrijving",
		    "filters": [
		      {
		        "name": "gras",
		        "count": 33,
		        "id": 2592
		      }
		    ]
		  }
		]
		
		GET: /entity/<id>
		standaard getById
		
		GET: /entity/<id>/reference
		overzicht van alle OneToMany relaties met de bijbehorende aantallen
		voorbeeld: /onderzoek/2951/meetpunt/2952/reference
		
		[
		  {
		    "name": "bodemlagen",
		    "type": "bodemlaag",
		    "count": 0
		  },
		  {
		    "name": "filters",
		    "type": "meetpuntfilter",
		    "count": 1
		  },
		  {
		    "name": "bodemmonsters",
		    "type": "bodemmonster",
		    "count": 0
		  },
		  {
		    "name": "afwerkingen",
		    "type": "meetpuntafwerking",
		    "count": 0
		  },
		  {
		    "name": "casingen",
		    "type": "casing",
		    "count": 0
		  }
		]
		
		GET: /entity/filterinfo
		Een overzich van alle specifiek gedefinieerde filters, op dit moment standaard op basis van reflectie, niet alle filters werken nog
		
		GET: /entity/writerinfo
		Welke writer gebruikt wordt voor de entiteit en of deze expansions ondersteunt (=debug informatie)
		
		POST: /entity
		create new instance
		
		POST: /entity/<id>
		updates instance
		
		DELETE: /entity/<id>
		removes instance
		
		
		/session
		
		GET: /session
		geeft lijst van actieve sessies terug (altijd maar 1)
		
		POST: /session
		login: {username: "", password: ""}
		geeft een id (uuid) terug waarmee de sessie uniek geidentificeerd kan worden
		
		GET: /session/<id>
		geeft informatie terug over de actieve sessie
		
		DELETE: /session/<id>
		logout
*/
	/*
		Heren, goedemorgen/-middag,
		
		Ik heb zojuist een update gedaan van office.
		
		De wijzigingen:
		- een onderzoek heeft nu een veld 'contour' waar WKT in kan (@karel: polygonen in tekstformaat)
		- de /query interface op de rest service heeft een extra veld 'filterExpression' gekregen, waarmee AND/OR constructies kunnen worden gemaakt.
		
		Zoals afgesproken, het formaat is, bijv: 
		["and", ["contains", "naam", "test"], ["contains", "projectcode", "1"]] (uiteraard zijn meerdere AND/OR's toegestaan) 
		of, een simpel voorbeeld:
		["contains", "naam", "test"]
		
		Om compabiliteitsredenen is het veld 'filters' gebleven. De standaard operator voor alle filters in dit veld is AND. Je kunt filters en filterExpression afzonderlijk of samen gebruiken. Igv. samen is de werking:
		filters AND filterExpression
		
		De operatoren anders dan AND en OR zijn dezelfde zoals je die gebruikt bij de filters, bijv: contains, equals, .... Ik zag dat ik NOT EQUALS nog niet bestond dus die heb ik toegevoegd.
		
		Voor zover ik weet wordt /query alleen in de rapportages gebruikt, dus kijk extra goed of e.e.a. daar nog naar behoren werkt. Ik heb wat profielen gegenereerd en het zag er op het eerste gezicht goed uit, maar what do i know ;)
		
		Groeten,
		Arnoud
	*/
	
