define(function (require) {
	
	/* 2021/01/23: new models */

	/* 2018-01-20: enums not to be treated like references */

    var Session = require("veldoffice/Session");
    var models = require("veldoffice/models");
    
    require("veldoffice/nameOf");
    
    var EM, ROOT = "/office-rest/rest/v1/persistence";
    var ROOT_WALK = "/office-rest/action/rapportage";
    var instances = {};
    var walks = {};
    
    function process(EM, entity, attributes, path, result) {
    	if(!attributes || attributes.length === 0) {
    		var key = path.split("/").pop().toLowerCase();
    		result = result.map(_ => _[key]);
    	}
    	return EM.processQueryResult(entity, path, result);
	}
	function getQueryPostData(path, attributes, criteria) {
		/*- Copy the criteria object */
    	var r = {};//js.mixIn(criteria);
    	
    	/*- Allow usage like: Query.execute({attributes: "type,count:id count"}) */
    	if(typeof attributes === "string") {
    		attributes = attributes.split(",");
    	} 
    	
		r.fields = {};
		(attributes || []).forEach(function(a) {
			a = a.split(" ");
			r.fields[a[a.length === 2 ? 1 : 0]] = a.shift();
		});
    	
    	if(criteria && criteria.where) {
    		r.filterExpression = criteria.where;
    	}
    	
    	if(criteria && criteria.having) {
    		r.having = criteria.having;
    	}
    	
    	/*- Convert string usage */
    	if(criteria && typeof criteria.groupBy === "string") {
    		r.groupBy = criteria.groupBy.split(",");
    	}
    	
    	/* Convert string usage */
    	if(criteria && typeof criteria.orderBy === "string") {
    		r.orderBy = criteria.orderBy.split(",").map(function(ob) {
    			ob = ob.split(" ");
    			return { path: ob[0], type: ob[1] || "asc" };
    		});
    	}
    	
    	if(criteria && criteria.hasOwnProperty("page")) {
    		if((r.page = criteria.page) instanceof Array) {
	    		r.pagesize = r.page[1] || 50;
	    		r.page = r.page[0];
    		} else {
    			r.pagesize = 50;
    		}
    	} else {
			r.pagesize = 50;
    	}
    	
    	if(criteria && criteria.hasOwnProperty("pagesize")) {
    		r.pagesize = criteria.pagesize;
    	}
    	
    	r.page = parseInt(r.page || 0, 10);
    	r.pagesize = parseInt(r.pagesize, 10);
    	
    	return JSON.stringify(r);
	}
	function getEntity(instance) {
		return instance._entity || instance['@entity'];
	}
	function getPath(instance) {
		return instance._path || instance['@path'];
	}
	function getKey(instance) {
		return instance._values.id || instance['@key'];
	}
	function getValues(instance) {
		// get values to post?
	}
	
	function allButId(obj) {
		obj = js.mixIn(obj);
		delete obj.id;
		return obj;
	}
	function generateId() {
		return Math.random().toString(36).substring(3, 5) + Date.now().toString(36);
	}
	
	var Factories = {
		AUTO_INC: Date.now(), // autogen'd ids are easily id'd (pun intended)
		
		instance: function(entity, path, values) {
			if(!instances[entity]) instances[entity] = {};
			
			if(!values.hasOwnProperty("id")) {
				if(isNaN(values.id = parseInt(path.split("/").pop(), 10))) {
					console.log("no id", entity, path, values);
					values.id = this.AUTO_INC++;
				}
			}
			
			var instance = instances[entity][values.id];
			var model = models[entity];
			if(model) {
				for(var k in values) {
					if(model[k] && model[k].type === "double" && values[k] === "null") {
						console.log(k, "double => null", entity);
						values[k] = null;
					}
				}
			}
			if(!instance) {
				instance = (instances[entity][values.id] = Object.create({
					_path: path,
					_prototype: values,
					_entity: entity,
					_values: Object.create(values, {})
				}, this.defineProperties(path, model)));
				/*- TODO Trigger event */
			} else {
				// console.log(entity, path, values);
				js.mixIn(instance._prototype, values);
				/*- TODO Trigger event */
			}

			if(model) {			
				/*- Copy non-model properties (usually agg funcs, joined attributes or specifc getters (calculated)) */
				for(var k in values) {
					if(!model.hasOwnProperty(k)) {
						var names = k.split(".");
						var a = names.pop();

						if(js.get(names.join(".model.") + ".model." + a, model)) {
						// `#VA-20200811-1` this is a joined attribute  (resolved 2021-08-09, extended 2024-10-07)
							var ninstance = instance;
							var nmodel = model, npath = [];
							names.forEach((name, i) => {
								npath.push(name);
								nmodel = nmodel[name];
								// if(ninstance[name] === undefined) {
								var key = values[npath.join(".") + ".id"];
								ninstance[name] = Factories.instance(
									nmodel.model_, 
									path + "?", 
									js.get(names.slice(0, i - 1).join("."), values) 
									|| {id: key});
								// }
								ninstance = ninstance[name];
								nmodel = nmodel.model;
							});
							ninstance[a] = values[k];
						} else {
							// console.log("not found", k, model);
							instance[k] = values[k];
						}
					}
				}
			}
			
			return instance;
		},
		set: {
			getter: function(path_, model, property, asIs) {
				return function() {
					if(EM.autoRequest !== true || this._prototype.hasOwnProperty(property)) {
						return this._values[property];
					}
				
					var me = this, attr = model[property];
					var path = path_ + "/" + this._prototype.id + "/" + (attr.path || attr.model_.toLowerCase());
				
					if(attr.model_.toLowerCase() === "veldwaarneming") {
						path = path.replace("bodemmonster", "veldmonster").replace("watermonster", "veldmonster");
						console.log(path + " adjusted");
					}

					return (this._prototype[property] = Session.get(this.ROOT + "/" + path))
						.then(function(resp) {
							return (me._prototype[property] = 
								asIs ? resp :
								resp.map(_ => Factories.instance(
									attr.model_, path + "/" + _.id, _)));
						});
				};
			},
			setter: function(path, model, property) {
				return function(value) {
					if(this._values[property] !== value) {
						this._values[property] = value;
						EM.changed(this, property);
					}
				};
			}
		},
		ref: {
			getter: function(path_, model, property, asIs) {
				return function() {
					var value;
					if(this._values.hasOwnProperty(property) || this._prototype.hasOwnProperty(property)) {
						value = this._values[property];
					} else {
						value = model[property].defaultValue;
					}
					if(value !== null && value !== undefined) {
						var me = this, attr = model[property];
						var path = path_ + "/" + (attr.path || attr.model_.toLowerCase());
						
						value = EM.get(attr.model_, value.id); // TODO determine key (not always id is it?)
					}
					return value;// === undefined ? null : value;
				};
			},
			setter: function(path, model, property) {
				return function(value) {
					if(this._values[property] !== value) {
						this._values[property] = value;
						EM.changed(this, property);
					}
				};
			}
		},
		has: {},
		'enum': {
			getter: function(path_, model, property, asIs) {
				return function() {
					if(this._values.hasOwnProperty(property)) {
						return this._values[property].name;
					} else if(this._prototype.hasOwnProperty(property)) {
						return this._prototype[property].name;
					}
					return model[property].defaultValue;
				};
			},
			setter: function(path, model, property) {
				return function(value) {
					if(this._values[property] !== value) {
						this._values[property] = value;
						EM.changed(this, property);
					}
				};
			}
		},
		value: {
			getter: function(path, model, property) {
				return function() {
					if(this._values.hasOwnProperty(property)) {
						return this._values[property];
					} else if(this._prototype.hasOwnProperty(property)) {
						return this._prototype[property];
					}
					return model[property].defaultValue;// || null;
				};
			},
			setter: function(path, model, property) {
				return function(value) {
					if(this._values[property] !== value) {
						this._values[property] = value;
						EM.changed(this, property);
					}
				};
			}
		},
		func: {
			getter: function(path, model, property) {
				return function() {
					return model[property].apply(this, arguments);
				};
			},
			setter: function(path, model, property) {
				return function(value) {
					// console.warn("readonly", model, property);
					try {
						return model[property].apply(this, arguments);
					} catch(e) {
						// EM.changed(this, property);
					}
				};
			}
		},
		defineProperties: function(path, model) {
			var properties = {};
			for(var k in model) {
				var prop = properties[k] = {
				    enumerable: true,
				    writeable: false,
				    configurable: true
				};
				var attr = model[k];

				if(typeof attr === "function") {
					prop.get = Factories.func.getter(path, model, k, attr.asIs);
					prop.set = Factories.func.setter(path, model, k);
					continue;
				}
				
				switch(attr.type) {
					case "set": 
						prop.get = Factories.set.getter(path, model, k, attr.asIs);
						prop.set = Factories.set.setter(path, model, k);
						break;
						
					case "ref": 
						prop.get = Factories.ref.getter(path, model, k, attr.asIs);
						prop.set = Factories.ref.setter(path, model, k);
						break;

					default:
						prop.get = Factories.value.getter(path, model, k, attr.asIs);
						prop.set = Factories.value.setter(path, model, k);
						break;
				}
			}
			
			return properties;
		}
	};
    			
    var infoOf = (e) => Object.keys(EM.instances[e]).map(_ => EM.info(EM.instances[e][_]));

    return (EM = {
    	ROOT: ROOT,
    	ROOT_WALK: ROOT_WALK,
    	
    	walks: walks,
    	instances: instances,
    	models: models,
    	
    	entityByPath: {
    		"veldoffice/bedrijf": "Bedrijf",
    		"veldoffice/account": "Account",
			"veldoffice/sessie": "Sessie",
			"veldoffice/onderzoek": "Onderzoek",
			"veldoffice/meetpunt": "Meetpunt",

			"veldoffice/aanleiding": "Aanleiding",
			"veldoffice/afwerking": "Afwerking",
			"veldoffice/asbest": "Asbest",
			"veldoffice/bedrijvenregelingstatustype": "Bedrijvenregelingstatustype",
			"veldoffice/bijzonderheid": "Bijzonderheid",
			"veldoffice/bijzonderheidgradatie": "Bijzonderheidgradatie",
			"veldoffice/boorapparaat": "Boorapparaat",
			"veldoffice/bronhouder": "Bronhouder",
			"veldoffice/eenheid": "Eenheid",
			"veldoffice/exmonsterverpakking": "ExMonsterverpakking",
			"veldoffice/fase": "Fase",
			"veldoffice/geslacht": "Geslacht",
			"veldoffice/geur": "Geur",
			"veldoffice/geurint": "Geurint",
			"veldoffice/grind": "Grind",
			"veldoffice/grondsoort": "Grondsoort",
			"veldoffice/grondsoortmatrix": "Grondsoortmatrix",
			"veldoffice/hoedanigheid": "Hoedanigheid",
			"veldoffice/humus": "Humus",
			"veldoffice/kenmerk": "Kenmerk",
			"veldoffice/kleur": "Kleur",
			"veldoffice/kleurint": "Kleurint",
			"veldoffice/laagsoort": "Laagsoort",
			"veldoffice/lovcertificering": "Lovcertificering",
			"veldoffice/lovcomponent": "Lovcomponent",
			"veldoffice/loveenheid": "Loveenheid",
			"veldoffice/lovmonstertype": "Lovmonstertype",
			"veldoffice/lovreferentie": "Lovreferentie",
			"veldoffice/lutum": "Lutum",
			"veldoffice/materiaalaanv": "Materiaalaanv",
			"veldoffice/materiaalafd": "Materiaalafd",
			"veldoffice/mediaan": "Mediaan",
			"veldoffice/meetpunttype": "MeetpuntType",
			"veldoffice/monstersoort": "Monstersoort",
			"veldoffice/oliewatermate": "Oliewatermate",
			"veldoffice/ondtype": "Ondtype",
			"veldoffice/oorsprong": "Oorsprong",
			"veldoffice/referentietype": "Referentietype",
			"veldoffice/schuifsterkte": "Schuifsterkte",
			"veldoffice/situatiebeschrijving": "Situatiebeschrijving",
			"veldoffice/steensoort": "Steensoort",
			"veldoffice/subjecttype": "SubjectType",
			"veldoffice/tank": "Tank",
			"veldoffice/veiligheidsklasse": "Veiligheidsklasse",
			"veldoffice/veldmatrix": "Veldmatrix",
			"veldoffice/veldwaarnemingtype": "VeldwaarnemingType",
			"veldoffice/vervolg": "Vervolg",
			
			"veldoffice/exbelucht": "ExBelucht",
			"veldoffice/exbemonsteringsmethode": "ExBemonsteringsmethode",
			"veldoffice/exdoorloop": "ExDoorloop",
			"veldoffice/exhelderheid": "ExHelderheid",
			"veldoffice/expeilbuismateriaal": "ExPeilbuismateriaal",
			"veldoffice/expompmethode": "ExPompmethode",
			"veldoffice/exgefiltreerd": "ExGefiltreerd",
			"veldoffice/exonderzoeksmethode": "ExOnderzoeksmethode",
			"veldoffice/exsteensoort": "ExSteensoort",
			"veldoffice/exschuifsterkte": "ExSchuifsterkte"
    	},

    	entities: function() {
    		return Object.keys(VO.em.instances).reduce((a, k) => { 
				a[k] = Object.keys(VO.em.instances[k]).map(_ => VO.em.instances[k][_]); 
				return a; 
			}, {});
    	},
    	info: function(obj, key) {
    		if(typeof obj === "string") {
    			if(key !== undefined) {
    				return EM.info(EM.get(obj, key));
    			}
    			return EM.instances[obj] && infoOf(obj);
    		}

    		var r = {}; 
    		"entity,values,prototype,path".split(",").map(_ => r[_] = obj['_' + _]); 
    		
    		// r.entity_ = r.entity;	
    		r.model = EM.models[r.entity];
    		r.instance = obj;
    		return r;
    	},
		isDirty: function(instance, attribute) {
			if(typeof attribute === "string" && attribute.indexOf(".") !== -1) {
				attribute = attribute.split(".");
				instance = instance[attribute.shift()];
				return instance && EM.isDirty(instance, attribute.join("."));
			}

			var values = js.mixIn({}, instance._values, true);
			var model = this.models[instance._entity];
			for(var k in values) {
				var m = model[k];
				if(!model.hasOwnProperty(k) || instance._prototype[k] === values[k]) {
					if(m && m.type === "has" && EM.isDirty(values[k])) {
						//
					} else {
						delete values[k];
					}
				}
			}

			var keys = Object.keys(values);
			return attribute === undefined ? keys.length > 0 : keys.indexOf(attribute) !== -1;
		},
		changed: function(instance, property) {
			// console.log("EM.changed", EM.info(instance), property);
		},

		newInstance: (entity, values) => Factories.instance(entity, "", js.mi({id: `.${generateId()}`}, values)),
		
		instance: function(entity, key) {
			// returns a entities/Instance for the calling instance/entity key
			
			if(arguments.length === 1) {
				// Called either as (instance) or (entity, key)
				return this.toInstance(entity);
			}
			return this.toInstance(this.get(entity, key));
		},
		toInstance: function(instance) {
			var Instance = require("entities/Instance");

			if(instance instanceof Instance) return instance;
			// if(typeof instance === "string" && arguments.length === 2) {
			// 	var current = EM.get(instance._entity, values.id);
			// 	return new Instance(instance, arguments[1], )
			// }
			if(instance && instance._values && instance._entity && instance._prototype) {

				var proto = js.trim(js.mi(instance._prototype));
				var values = js.trim(js.mi(instance._values));
				
				var current = this.get(instance._entity, proto.id || values.id);
				
				js.mixIn(current._prototype, proto);
				js.mixIn(current._prototype, values);
				js.mixIn(current._values, values);
				
				var r = new Instance(instance._entity, proto.id, current._prototype);

				return r;
			}
		},

    	get_: function(entity, key) {
    		// TODO refector to fetch?
			var path = entity.toLowerCase().split(":").pop();
        	return Session.get(String.format("%s/%s/%s", this.ROOT, path, key))
        		.then(function(result) {
        			return Factories.instance(entity, path + "/" + key, result);
	        	});
    	},
    	get: function(entity, key) {
			var path = entity.toLowerCase().split(":").pop();
			
			// if(!instances[entity] || !instances[entity][key]) {
				return Factories.instance(entity, path + "/" + key, {id: key});
			// }
			// return instances[entity][key];
    	},
    	
    	clear: function(instance, valuesOnly, skip) {
    		if(arguments.length === 2 && valuesOnly instanceof Array) {
    			skip = valuesOnly;
    			valuesOnly = false;
    		} else {
    			skip = skip || [];
    		}
    		for(var k in instance._values) {
    			if(skip.includes(k)) continue;
    			delete instance._values[k];
    		}
    		if(valuesOnly !== true) {
    			for(var k in instance._prototype) {
	    			if(skip.includes(k) || k === "id") continue;
    				delete instance._prototype[k];
    			}
    			EM.changed(instance, ".cleared");
    		}
    		return instance;
    	},
    	revert: function(instance, attribute) {
    		if(arguments.length === 2 && typeof attribute === "string") {
    			if(instance._values.hasOwnProperty(attribute)) {
	    			delete instance._values[attribute];
	    			EM.changed(instance, attribute)
	    			return true;
    			}
    		} else if(arguments.length === 1) {
    			var keys = Object.keys(instance._values);
    			keys.forEach(attribute => {
    				delete instance._values[attribute]
    			});
    			EM.changed(instance, keys);
    			return true;
    		}
			return false;
    	},

		remove: function(instance) {
    		var entity = getEntity(instance);
    		var path = getPath(instance);
    		var key = getKey(instance);

    		return Session['delete']({
	    			url: String.format("%s/%s", this.ROOT, path), 
			        contentType: "application/json"
	    		});
			// return js.waitAll("255ms");
		},
		update: function (instance, values) {
		    var entity = getEntity(instance);
		    var path = getPath(instance);
		    var key = getKey(instance);
		
		    values = values || js.mixIn({ id: key }, instance._values);
		
		    return Session.post({
		        url: String.format("%s/%s", this.ROOT, path),
		        contentType: "application/json",
		        data: JSON.stringify(values)
		    });
		},
		
    	persist: function(instance, opts) {
    		opts = opts || {};
    		// var entity = getEntity(instance);
    		var path = opts.path || getPath(instance);
    		var key = getKey(instance);
    		
    		if(typeof key === "string" && key.startsWith(".")) {
    			// new instance, should catch new server id
    		}
    		
    		values = opts.values || js.mixIn({id: key}, instance._values);
    		
    		return Session.post({
    			url: opts.url || js.sf("%s/%s", this.ROOT, path), 
		        contentType: "application/json",
		        data: JSON.stringify(values)
    		});
    	},
		query_: function(entity, attributes, criteria, opts) {
			var path = (opts && opts.path) || entity.toLowerCase().split(":").pop();
				criteria = js.mixIn(criteria);
    		var count_criteria = criteria.count ? 
	    			criteria.count === true ? criteria : criteria.count
	    		: false;
    		
			delete criteria.count;
			
			opts = js.mixIn(opts || {});
			
			if(attributes === "") attributes = [];
			
			if(entity.indexOf(":") !== 1) entity = entity.split(":")[0];
        	
        	if(count_criteria) {
    			return EM.count(path, count_criteria).then(
					(count) => EM.query(entity, attributes, criteria, opts).then(
    					(objs) => [count, objs, count_criteria.result]));
        	}

		    return Session.post({
			        url: String.format("%s/%s/query", opts.ROOT || this.ROOT, path),
			        contentType: "application/json",
			        data: getQueryPostData(path, attributes, criteria)
			    }).then(function(result) {
			    	return opts.raw === true ? result : process(EM, entity, attributes, path, result);
			    });
        },
    	query: function(entity, attributes, criteria, opts) {
    		if(attributes === "*") {
    			attributes = [];
    			// Object.entries(models[entity])
    			// 	.filter(e => e[1].type !== "set")
    			// 	.filter(e => e[1].type !== "has")
    			// 	.filter(e => e[1].type !== "ref")
    			// 	.map(e => e[1].type === "has" || e[1].type === "ref" ? "[outer]" + e[0] : e[0])
    		}
    		
    		if(typeof attributes === "object" && !(attributes instanceof Array) && arguments.length >= 2) {
    			opts = criteria;
    			criteria = attributes;
    			attributes = "";
    		} else if(arguments.length === 1 && typeof entity === "object" && typeof entity.path === "string") {
    			return this.query_(entity.path, entity.attributes, entity.criteria, entity.opts);
    		}

    		if(js.get("page.length", criteria) === 3) {
    			opts = opts || {};
    			
    			var path = entity.toLowerCase().split(":").pop();
    			entity = entity.split(":")[0];
    			
    			var pager = criteria.page[2];
    			var raw = opts.raw === true || (criteria.raw === true);
    			opts.raw = true;

    			var me = this;
    			return pager.load(criteria.page[0]).then(function(page) {
    				if(!page.data) {
    					// console.log("EM.query (requesting page)", page._id);
    					return me.query_.apply(me, [entity, attributes, criteria, opts]).then(function(res) {
    						page.data = res;
    						return pager.save(page).then(function() {
						    	return raw ? res : process(EM, entity, attributes, path, res);
    						});
    					});	
    				} else {
    					// console.log("EM.query (no request, using cached page)", page);
				    	return raw ? page.data : process(EM, entity, attributes, 
				    		path, page.data);
    				}
    				return page.data;
    			});
    		}
    		
    		if(opts === undefined) {
    			opts = {};
    			if(criteria) ["raw", "ROOT"].forEach(k => {
    				if(criteria.hasOwnProperty(k)) {
		    			opts[k] = criteria[k];
		    			delete criteria[k];
    				}
    			});
    		}
    		
    		return this.query_(entity, attributes, criteria, opts);
    	},
    	count: function(path, criteria) {
    		var attrs = (criteria||{}).attributes || "count:id";
    		var count_key = attrs.split(/,|\s/)[0];
    		var opts = {raw: true};
    		if(!(opts.ROOT = (js.get("options.ROOT", criteria) || js.get("opts.ROOT", criteria)))) {
    			delete opts.ROOT;
    		}
    		// var attrs = "count:id";
    		return this.query(criteria.path || criteria.entity || path, attrs, criteria, opts)
    			.then(function(res) {
    				// return res[0][attrs.split(" ").pop()];
    				criteria.result = res;
    				return (res[0] || {})[count_key];
    			});
    	},

    	walk: function(entity, key, view, opts) {
			/*- Y dice: walk remote (if not cached) and then walk local -*/

    		opts = opts || {/* raw: false */};
			var cacheKey = opts.cacheKey || js.sf("%s.%s.%s", entity, key, view);
    		
			var cacheEntry = this.walks[cacheKey];
    		if(cacheEntry instanceof Promise && opts.useCache !== false) {
    			return cacheEntry;
    		}
    		
			var process = (res) => {
				if(opts === true || (opts !== undefined && opts.raw === true) || typeof res === "string") {
					return res;
				}
				// if(typeof res.instances[entity][key] !== "object") {
				// 	try {
						
						console.log("process-walk-result", this.processWalkResult2(res));
				// 	} catch(e) {
				// 		console.error(entity, key, res, e);
				// 		throw e;
				// 	}
				// }
				
				return this.get(entity, key);
			}
			
			if(opts.updateCacheEntryWith) {
				return (this.walks[cacheKey] = new Promise(resolve => resolve(opts.updateCacheEntryWith)));
			}

    		key = window.escape(key);
    		view = window.escape(view);
    		
    		if(entity === "") {
	    		return (this.walks[cacheKey] = Session.get(
	    			js.sf("%s?view=%s&id=%s%s", this.ROOT_WALK, view, key, opts.test ? "&test=" + opts.test : ""))
	    			.then(res => process(res)));
    		}
    		
    		return (this.walks[cacheKey] = Session.get(
    			js.sf("%s?view=%s&entity=%s&id=%s%s", this.ROOT_WALK, view, entity, key, opts.test ? "&test=" + opts.test : ""))
    			.then(res => process(res)));
    	},
    	
		join: (terms, op) => terms.length <= 1 ? terms.pop() : [op, terms.pop(), EM.join(terms, op)],
		and: (terms) => EM.join(terms, "and"),
		or: (terms) => EM.join(terms, "or"),
		
    	fixReferences: function(entity, instance) {
    		/*- !! Modifies instance, more precise, it resolves references */
    		if(instance.id && !isNaN(instance.id)) {
    			instance.id = parseInt(instance.id, 10);
    		}
    		
			var model = models[entity];
			for(var k in model) {
				if(instance.hasOwnProperty(k)) {
					var attr = model[k], value = instance[k];
					if(attr.model) {
						if(value !== null && value !== undefined && (attr.type === "ref" || attr.type === "has") && !value._entity) {
							// VO: no instance/_entity -> id can string or number
							value = Factories.instance(attr.model_, entity + "/" + k, {id: value.id || value});
						} else if(attr.type === "set") {
							value = (value||[]).map(instance => {
								if(instance && !instance._entity) {
									// VO: no instance/_entity -> id can string or number
									return Factories.instance(attr.model_, entity + "/" + k, {id: instance.id || instance});
								}
							});
						}
						instance[k] = value;
					}
				}
			}
			
			return instance;
    	},
		reduce_or: function(acc, current, index, array) {
			return index ? ["or", current, acc] : current;
		},
		processCoderingInstances: function(instances, path) {
			var entity = this.entityByPath["veldoffice/" + path];
			var ctx = js.ctx(arguments.callee);
			
			if(!entity) entity = this.entityByPath["veldoffice/" + path.toLowerCase()];
			if(!entity) throw new Error("Unknown entity: " + path);
			
			if(!ctx.hasOwnProperty(entity)) {
				ctx[entity] = true;
				for(var k in instances[path]) {
					if(instances[path].hasOwnProperty(k)) {
						var _ = instances[path][k];
						// _.vid = _.id;
						// _.id = _.xid;
						Factories.instance(entity, path + "/" + _.id, _);
					}
				}
			} else {
				// console.log("already processed", entity)
			}
		},
		processQueryResult: function(entity, path, result) {
			if(!(result instanceof Array)) {
				throw new Error("Unexpected result");
			}
			return result.map(_ => Factories.instance(entity, path + "/" + _.id, _));
		},
		processWalkResult: function(result) {
			result.objs.forEach(obj => {
				for(var k in obj) {
					if(obj[k] instanceof Array) {
						if(obj[k].length > 0) {
							var elem0 = obj[k][0];
							if(elem0 instanceof Array) {
								// resolve an array of references
								obj[k] = obj[k].map(_ => result.objs[_]);
							} else {
								// resolve single reference (TODO maybe attribute is Array-like)
								obj[k] = result.objs[elem0];
							}
						} else {
							// empty set, no change necessary
						}
					}
				}
			});
			for(var entity in result.instances) {
				for(var key in result.instances[entity]) {
					var index = result.instances[entity][key];
					result.instances[entity][key] = result.objs[index];
					// instances[entity][key]['@key'] = key;
				}
			}
			return result;
		},
		processWalkResult2: function(result) {
			var vm = result.instances.Veldmonster;
			var wm = result.instances.Watermonster || {};
			var bm = result.instances.Bodemmonster || {};

			for(var k in vm) {
				if(wm[k]) {
					js.mixIn(result.objs[wm[k]], result.objs[vm[k]]);
				} else if(bm[k]) {
					js.mixIn(result.objs[bm[k]], result.objs[vm[k]]);
				}
			}
			
			// process with Factories => EM.instances
			result.objs.forEach(obj => {
				if(obj.geoTechnisch) {
					var value, gt = result.objs[obj.geoTechnisch[0]];
					for(var gtk in gt) {
						if(	['@', '_'].includes(gtk.charAt(0)) === false && 
							["id", "created", "modified"].includes(gtk) === false
						) {
							if((value = gt[gtk]) !== undefined) { // when geoTechnisch-value 
								if(value instanceof Array && value.length === 1) {
									value = result.objs[value[0]]; // reference to brobhrgt?
									if((value = value.naam) !== undefined) {  // naam as id? TODO exportid-needed for sikb-9.1.0
										value = value.split(" ")[0];
									}
								}
								if(value !== undefined && (obj[js.sf("brobhrgt:%s", gtk.toLowerCase())] !== value)) {
									// console.log("gt-value-override", gtk, value, obj[js.sf("brobhrgt:%s", gtk.toLowerCase())]);
									obj[js.sf("brobhrgt:%s", gtk.toLowerCase())] = value;
								} else {
									// console.log("gt-value-equals", gtk, value);
								}
							}
						}
					}
					// console.log("gt-obj", gt, obj);
				}
				for(var k in obj) {
					if(obj[k] instanceof Array) {
						if(obj[k].length > 0) {
							var elem0 = obj[k][0];
							if(elem0 instanceof Array) {
								if(elem0[0] === this) { // special case-hack
									// static value, do nothing but get rid of first elem
									elem0.splice(0, 1);
									obj[k] = elem0;
									// console.log(k, obj[k]);
								} else {
									// resolve an array of references
									obj[k] = obj[k].map(_ => result.objs[_]);
								}
							} else {
								// resolve single reference (TODO maybe attribute is Array-like)
								obj[k] = result.objs[elem0];
							}
						} else {
							// empty set, no change necessary
						}
					}
				}
			});
			
			for(var entity in result.instances) {
				var map = result.instances[entity];
				for(var key in map) {
					var instance = result.objs[map[key]];
					EM.fixReferences(entity, instance);
					
					if(entity === "Onderzoek") {
						if(typeof instance.status === "string") {
							instance.status = { id: instance.status, 
								name: locale("Onderzoek-status/." + instance.status) };
						}
					}
					
					Factories.instance(entity, entity + "/" + key, instance);
				}
			}

			for(var entity in result.instances) {
				for(var key in result.instances[entity]) {
					var index = result.instances[entity][key];
					result.instances[entity][key] = result.objs[index];
					// instances[entity][key]['@key'] = key;
				}
			}

			return result;
		},
		processChanges: function(instance, changes) {
			// debugger;
			var info = this.info(instance);
			if(changes.values && instance.id === ".new") {
				if(changes.values.proto) {
					js.mixIn(instance._prototype, EM.fixReferences(info.entity, changes.values.proto));
					delete changes.values.proto;
				}
			}
			if(changes.values) {
				js.mixIn(instance._values, EM.fixReferences(info.entity, changes.values));
			}
			console.log("EM.processChanges", changes, info);
		}
    });
});

	/*-
		Hey man,
		
		De applicatie is nu in 2'en gedeeld:
		1) de back-end (het oude veldoffice project, development deploy versie meegeleverd)
		2) de front-end (source-versie): veldoffice2-front-end, zie svn
		
		In principe is de front-end qua inhoud hetzelfde, alleen bevat het nu bower en grunt configuraties. De hele bende wordt samengepakt etc. maar volgens mij ben je er wel mee bekend, dus als je de gruntfile.js en bower.json bekijkt zullen de details wel duidelijk worden. Het oude office-front-end project wordt vanaf nu niet meer gebruikt.
		
		Voor de profielen: omdat het aanzwengelen van het script in de productieversie niet zou werken (harde referentie naar c:/data/ ...) heb ik er een setting van gemaakt. In de back-end war zit een development.settings.properties (veldoffice-ejb/config) die je kunt aanpassen om zo je lokale script aan te roepen:
		settings.development.properties -> profielen.script=c:/data/office/profielen/run.js
		
		Op de back-end is nu ook een query interface beschikbaar: 
		POST: /<entity>/query of /<entity1>/<entity2>/.../<entityn>/query
		
		Post een object met het volgende formaat:
		- fields: array van de te selecteren velden: alias, path, bijv. { "pc": "projectcode", "mp": "countDistinct:meetpunten"}
		- filters: array van filter criteria, { path, type, value }: { path: "bodemlagen.onderkant", type: "gte", value: 10}
		- orderBy: array van sorteervolgordes, { path, type (desc, asc)} bijv. {path: "meetpunten.created", type: "desc"}, asc is default
		- groupBy: array van group by, bijv: meetpunten.type
		- having: array van expressies voor having van de vorm: <path><operator><value>, bijv. countDistinct:id>3
		- pagesize: paginagrootte
		- page: de op te vragen pagina
		
		filters types: contains, endswith, equals, gte, inlist, isempty, isnotnull, isnull, lt, lte, notequals, startswith
		aggregates: <func>:<path>, bijv (/onderzoek/query): countDistinct:meetpunten, max:meetpunten.bodemlagen.bovenkant
		
		Voor alle paden gelden de namen zoals gedefinieerd in het entity model. Onderzoek -> meetpunt -> bodemlaag is: 'meetpunten.bodemlagen'.
		
		Als je geen admin bent, wordt er automatisch gefilterd op bedrijf voor zover er een hierarchische relatie met de gevraagde entiteit bestaat.
		
		Simpel voorbeeldje:
		/onderzoek/query
		{
			fields: {
				"pc": "projectcode",
				"naam": "naam",
				"bodemlaag": "max:meetpunten.bodemlagen.onderkant"
			}, 
			groupBy: ["id"],
			orderBy: [{
					path: "naam",
					type: "desc"
			}],
			pagesize: 25,
			page: 3
		}
		
		Zie de front-end /test voor een kleine demo.
		
		Het is een eerste versie, dus het zou me niets verbazen wanneer er hier en daar nog wat bugs/fouten inzitten. Laat me weten hoe 't gaat.
		
		mzzls,
		Arnoud
	*/
	/*
		Standaard acties voor elke entiteit

		GET: /entity -> search
		
		optional parameters:
		pagesize: pagesize
		page: op te vragen pagina
		e: expansion 
		v: view -> voorgedefinieerd view (let op: views ondersteunen geen expansions)
		path=value -> selectiecriteria, voorbeelden: naam=vijfhuizen, tov.id=1234, ...
		sortby: pad om te sorteren, default = id
		sorttype: ASC, DESC, default = ASC
		
		GET: /entity/count
		Geeft het aantal terug, accepteert dezelfde filters als search
		
		
		GET: /entity/group
		overzicht van alle ManyToOne referenties van de selectie met de aantalen waarin ze voorkomen (gebruikt voor mogelijke filterwaarden in de grids)
		bijv. /onderzoek/2951/meetpunt/group
		[
		  {
		    "name": "tov",
		    "filters": [
		      {
		        "name": "NAP",
		        "count": 33,
		        "id": 2512
		      }
		    ]
		  },
		  {
		    "name": "type",
		    "filters": [
		      {
		        "name": "grondboring",
		        "count": 33,
		        "id": 2377
		      }
		    ]
		  },
		  {
		    "name": "situatiebeschrijving",
		    "filters": [
		      {
		        "name": "gras",
		        "count": 33,
		        "id": 2592
		      }
		    ]
		  }
		]
		
		GET: /entity/<id>
		standaard getById
		
		GET: /entity/<id>/reference
		overzicht van alle OneToMany relaties met de bijbehorende aantallen
		voorbeeld: /onderzoek/2951/meetpunt/2952/reference
		
		[
		  {
		    "name": "bodemlagen",
		    "type": "bodemlaag",
		    "count": 0
		  },
		  {
		    "name": "filters",
		    "type": "meetpuntfilter",
		    "count": 1
		  },
		  {
		    "name": "bodemmonsters",
		    "type": "bodemmonster",
		    "count": 0
		  },
		  {
		    "name": "afwerkingen",
		    "type": "meetpuntafwerking",
		    "count": 0
		  },
		  {
		    "name": "casingen",
		    "type": "casing",
		    "count": 0
		  }
		]
		
		GET: /entity/filterinfo
		Een overzich van alle specifiek gedefinieerde filters, op dit moment standaard op basis van reflectie, niet alle filters werken nog
		
		GET: /entity/writerinfo
		Welke writer gebruikt wordt voor de entiteit en of deze expansions ondersteunt (=debug informatie)
		
		POST: /entity
		create new instance
		
		POST: /entity/<id>
		updates instance
		
		DELETE: /entity/<id>
		removes instance
		
		
		/session
		
		GET: /session
		geeft lijst van actieve sessies terug (altijd maar 1)
		
		POST: /session
		login: {username: "", password: ""}
		geeft een id (uuid) terug waarmee de sessie uniek geidentificeerd kan worden
		
		GET: /session/<id>
		geeft informatie terug over de actieve sessie
		
		DELETE: /session/<id>
		logout
*/
	/*
		Heren, goedemorgen/-middag,
		
		Ik heb zojuist een update gedaan van office.
		
		De wijzigingen:
		- een onderzoek heeft nu een veld 'contour' waar WKT in kan (@karel: polygonen in tekstformaat)
		- de /query interface op de rest service heeft een extra veld 'filterExpression' gekregen, waarmee AND/OR constructies kunnen worden gemaakt.
		
		Zoals afgesproken, het formaat is, bijv: 
		["and", ["contains", "naam", "test"], ["contains", "projectcode", "1"]] (uiteraard zijn meerdere AND/OR's toegestaan) 
		of, een simpel voorbeeld:
		["contains", "naam", "test"]
		
		Om compabiliteitsredenen is het veld 'filters' gebleven. De standaard operator voor alle filters in dit veld is AND. Je kunt filters en filterExpression afzonderlijk of samen gebruiken. Igv. samen is de werking:
		filters AND filterExpression
		
		De operatoren anders dan AND en OR zijn dezelfde zoals je die gebruikt bij de filters, bijv: contains, equals, .... Ik zag dat ik NOT EQUALS nog niet bestond dus die heb ik toegevoegd.
		
		Voor zover ik weet wordt /query alleen in de rapportages gebruikt, dus kijk extra goed of e.e.a. daar nog naar behoren werkt. Ik heb wat profielen gegenereerd en het zag er op het eerste gezicht goed uit, maar what do i know ;)
		
		Groeten,
		Arnoud
	*/
