define(require => {

    var module = require("module");
    var locale = require("locale");

    var lastLogin = {};

    locale = locale.module(module) || window.locale;
    
    function set_hash(Session, value) {
    	Session.hash = value;
        lastLogin.timeout = Date.now() + 5 * 60 * 1000;
    }

    return {
        baseUrl: "/office-rest/rest/v1/",
        host: "",

        hash: undefined,
        ll: () => lastLogin,
        use_ll: true,

        needed() { /* stub for application, called by waitFor() */ },

        async refresh() {
            if (this.refreshPromise) return this.refreshPromise;

            this.refreshPromise = fetch(this.host + this.baseUrl + "session")
                .then(async (resp) => {

                    const data = await resp.json();
                    delete this.refreshPromise;

                    if (data.length) {
                    	set_hash(this, data[0]);
                        return data[0];
                    } else {
                    	set_hash(this);
                        return false; // not logged in
                    }
                })
                .catch((err) => {
                    console.error("Error during refresh", err);
                    throw err;
                });

            return this.refreshPromise;
        },
        async info(p_info = "") {
            if (this.infoPromise) return this.infoPromise;

            this.infoPromise = this.waitFor()
                .then(async (hash) => {
                    const url = `${this.host + this.baseUrl}session/${hash || this.hash}${p_info}`;
                    const response = await fetch(url);

                    if (!response.ok) throw new Error(`Info request failed: ${response.status}`);
                    
                    const data = await response.json();
                    this.details = data;
                    lastLogin = js.mi(lastLogin || {}, { timeout: Date.now() + 5 * 60 * 1000 });
                    return data;
                })
                .catch((err) => {
                    console.error("Error during info", err);
                    throw err;
                });

            return this.infoPromise;
        },
        async waitFor() {

        	if(this.hash && Date.now() > lastLogin.timeout) {
        		return new Promise(resolve => resolve(this.hash));
        	}

            return this.waitForPromise || 
            	(this.waitForPromise = new Promise((resolve, reject) => {
                	this.waitForPromise_resolve = resolve;
	            	this.refresh()
		                .then(() => {
		                    if (this.hash) {
		                    	resolve(this.hash);
		                    } else {
		                    	this.needed();
		                    }
		                })
		                .catch((err) => {
		                    console.error("Error during waitFor", err);
		                    reject(err);
		                });
	            }));
        },

        async login(user, pass) {
            lastLogin = { user, pass };

            const data = await this.post({url: "session/", body: { username: user, password: pass }}, false);

            if (data.success === true) {
	            set_hash(this, data.id);
	            lastLogin.success = true;
	            lastLogin.timeout = Date.now() + 5 * 60 * 1000;
	            
	            if(this.waitForPromise_resolve) {
	            	this.waitForPromise_resolve(this.hash);
	            	delete this.waitForPromise_resolve;
	            }
            }

            return data;
        },
        async logout(hash) {

        	hash = hash || this.hash;
            this.reset();
        	
        	if(!hash) return false;

            return await this.request({ url: `session/${hash}`, method: "DELETE", json: false }, false).then(() => true);
		},
		
        isAuthenticated() {
            return Boolean(this.hash);
        },
        reset() {
            delete this.infoPromise;
            delete this.waitForPromise;
            delete this.refreshPromise;
			delete this.details;
            delete this.hash;
        },

        async request({ url, method = "GET", body, headers = {}, data, 
        contentType = "application/json", json = true, waitFor = true }) {
        
            if(data) body = data;
            
        	if(waitFor === true) {
        		return this.waitFor().then(() => this.request({ 
        			url, method, body, headers, data, json, waitFor: false }));
        	}

            if (!url.startsWith("http")) {
                url = url.startsWith("/")
                    ? this.host + url
                    : this.host + this.baseUrl + url;
            }

            const response = await fetch(url, {
                body: typeof body === "string" ? body : JSON.stringify(body),
                headers: js.mi({ "Content-Type": contentType }, headers),
                method
            });

            if (!response.ok) {
            	throw new Error(`Request failed: ${response.status}`);
            }

            return json ? await response.json() : await response.text();
        },

		get_: function(what, waitFor = true) {
			console.log("deprecated, use request() instead");
            var d = new Deferred();
            this.available().addCallback((resp) => {
                ajax(this.host + this.baseUrl + what).then(function(resp) {
                    if (typeof resp === "string") {
                        resp = JSON.parse(resp);
                    }
                    d.callback(resp);
                }).fail(function(e) {
                    d.errback(e);
                });
                return resp;
            });
            return d;
		},
		get: function(what, waitFor = true) {
			if(typeof what !== "object") {
				what = {url: what};
			}
			what.method = "GET";
			what.waitFor = waitFor;
			return this.request(what);
		},
		put: function(what, waitFor = true) {
			if(typeof what !== "object") {
				what = {url: what};
			}
			what.method = "PUT";
			what.waitFor = waitFor;
			return this.request(what);
		},
		post: function(what, waitFor = true) {
			if(typeof what !== "object") {
				what = {url: what};
			}
			what.method = "POST";
			what.waitFor = waitFor;
			return this.request(what);
		},
		'delete': function(what, waitFor = true) {
			if(typeof what !== "object") {
				what = {url: what};
			}
			what.method = "DELETE";
			what.waitFor = waitFor;
			return this.request(what);
		},

        available() {
            console.log("deprecated, use waitFor() instead");
            return this.waitFor.apply(this, arguments);
        },
        isLoggedIn() {
            console.log("deprecated, use refresh() instead");
            return this.refresh();
        },
        invalidate() {
            console.log("deprecated, use refresh() instead");
        }
   
    };
});