define(function(require) {
	
	// var locale = window.locale;
	
	var Deferred = require("js/Deferred"), Session;
	var ajax = require("jquery").ajax;
	var module = require("module");
	var locale = require("locale");
	var js = require("js");
	
	var lastLogin;

	locale = locale.module(module) || window.locale;
	
	return (Session = {
		host: "",
		baseUrl: "/office-rest/rest/v1/",
		
		ll: () => { return lastLogin; },
		use_ll: true,

		available: function() {
			console.log("deprecated, use waitFor() instead");
			return this.waitFor.apply(this, arguments);
		},
		isLoggedIn: function() {
			console.log("deprecated, use refresh() instead");
			return this.refresh();
		},
		invalidate: function() {
			console.log("deprecated, use refresh() instead");
			// return this.available();
		},
		
		needed: function() { /* stub for application, called by waitFor() */ },

		refresh: function() {
			/* resolves to false when not logged in */
			if(this.refresh_ instanceof Deferred) {
				return this.refresh_;
			}
			if(this.login_ instanceof Deferred) {
				return this.login_;//.addCallback(() => this.refresh());
			}
			
			var d = new Deferred(), me = this;
			ajax(this.host + this.baseUrl + "session").then(function (resp) {
				if(d !== me.refresh_) {
					/*- Ignore callbacks from other sessions */
					return;
				}
				resp = typeof resp === "string" ? JSON.parse(resp) : resp;
				delete me.hash;
				delete me.refresh_;
				delete me.waitFor_;
				if(resp.length) {
					me.hash = resp[0];
					
					if(!lastLogin) {
						lastLogin = {timeout: Date.now() + 5 * 60 * 1000};
					}
					
					d.callback(resp[0]);
				} else {
					d.callback(false);//locale("-messages/#not-logged-in"));
				}
			});
			return (this.refresh_ = d);			
		},
		waitFor: function() {
			if((this.use_ll && lastLogin) && lastLogin.timeout < Date.now()) {
				delete lastLogin.timeout;
				delete this.waitFor_;
				delete this.hash;
			}
			
			if(this.waitFor_ instanceof Deferred) {
				// Session.needed();
				return this.waitFor_;
			}
			var d = new Deferred();
			
			if(this.hasOwnProperty("hash")) {
				d.callback(this.hash);
			} else {
				this.waitFor_ = d;
				ajax(this.host + this.baseUrl + "session").then(function (resp) {
					if(Session.waitFor_ !== d) {
						/*- Ignore callbacks from other sessions? */
						// throw new Error("Abort/Ignore");
					}
					resp = typeof resp === "string" ? JSON.parse(resp) : resp;
					if(resp.length) {
						Session.hash = resp[0];
						d.callback(Session.hash);
					} else {
						Session.needed();
					}
				}).fail(function(resp) {
					if(Session.waitFor_ !== d) {
						/*- Ignore callbacks from other sessions? */
						// throw new Error("Abort/Ignore");
					}
					d.errback(resp);					
				});
			}
			return d;
		},
		info: function(p_info) {
			if(this.info_ instanceof Deferred) {
				return this.info_;
			}
			var me = this, d = new Deferred();
			if(this.hasOwnProperty("info_")) {
				if(this.info_ instanceof Deferred) {
					// debugger;
					// console.log("!!!!")
				}
				d.callback(this.info_);
			} else {
				this.info_ = d;
				this.waitFor().addCallback(hash => {
					if(Session.info_ !== d) {
						/*- Ignore callbacks from other sessions */
						// d.cancel();
						return;
					}
					p_info = p_info || "";
					ajax(this.host + this.baseUrl + "session/" + (hash || this.hash) + p_info).then(function(resp) {
						Session.hash = resp.id;
						d.callback(Session.info_ = Session.details = resp);
						lastLogin = {timeout: Date.now() + 5 * 60 * 1000};
						return resp;
					});
					return hash;
				}).addErrback(function(e) {
					d.errback(e);
				});
			}
			return d;
		},
		
		get_: function(what, waitFor) {
			console.log("deprecated, use request() instead");
            var d = new Deferred();
            this.available().addCallback((resp) => {
                ajax(this.host + this.baseUrl + what).then(function(resp) {
                    if (typeof resp === "string") {
                        resp = JSON.parse(resp);
                    }
                    d.callback(resp);
                }).fail(function(e) {
                    d.errback(e);
                });
                return resp;
            });
            return d;
		},
		get: function(what, waitFor) {
			if(typeof what !== "object") {
				what = {url: what};
			}
			what.method = "GET";
			return this.request(what, waitFor);
		},
		put: function(what, waitFor) {
			if(typeof what !== "object") {
				what = {url: what};
			}
			what.method = "PUT";
			return this.request(what, waitFor);
		},
		post: function(what, waitFor) {
			if(typeof what !== "object") {
				what = {url: what};
			}
			what.method = "POST";
			return this.request(what, waitFor);
		},
		'delete': function(what, waitFor) {
			if(typeof what !== "object") {
				what = {url: what};
			}
			what.method = "DELETE";
			return this.request(what, waitFor);
		},
		request: function(what, waitFor) {
			var callee = arguments.callee;
			
			if(typeof what !== "object") {
				what = {url: what};
			}
			
			if(what.url.indexOf("http") !== 0) {
				if(what.url.charAt(0) === "/") {
					what.url = this.host + what.url;
				} else {
					what.url = this.host + this.baseUrl + "" + what.url;
				}
			}
			
			// TODO isPublic-API?
			if(what.url.indexOf("public") === -1 && this.use_ll && (!(what.ll || (what.ll = lastLogin && lastLogin.timeout > Date.now())))) {
				// make sure there is a session
				return this.refresh().then(loggedIn => {
					if(!loggedIn) {
						delete this.hash;
						
						// alert("making request and not logged in, logging in...");
						
						if(!lastLogin || (!lastLogin.user && !lastLogin.success)) {
							return Promise.resolve(this.needed()).then(res => {
								if(res) {
									what.ll = true;
									return this.request(what, waitFor);
								}
								throw new Error("No session available");// Session.no;
							});
						}
						return callee.login || (callee.login = this.login(lastLogin.user, lastLogin.pass)
							.then(res => {
								delete callee.login;
								what.ll = true;
								return this.request(what, waitFor);
							}));
					}
					lastLogin = js.mi(lastLogin || {}, { success: true, timeout: Date.now() + 5 * 60 * 1000 });
					return this.request(what, waitFor);
				});
			}
			
			var failed = false, auto_retry = true;
			return new Promise((resolve, reject) => {
				function request(resp) {
					ajax(what)
						.then(function(resp) {
							if(what.raw !== true && typeof resp === "string") {
								try {
									resp = JSON.parse(resp);
								} catch(e) {
									console.warn("unexpected response", {request: what, response: resp});
								}
							}
							resolve(resp);
						})
						.fail(function(e) {
							console.error("request", what, "failed with", e)
							// if(!failed && auto_retry !== false) {
								// failed = true;
								// retry 
							// } else {
								reject(e);
							// }
						});
						
					return resp;
				}
				
				if(waitFor !== false && what.url.indexOf("public") === -1) {
					return this.waitFor().addCallback(request);
				}
				request();
			});
		},
		
		isAuthenticated: function() {
			return this.hasOwnProperty("hash");
		},
		login: function(user, pass) {
			if(this.login_ instanceof Deferred) {
				console.warn("login called while logging in");
				return this.login_;
			} else if(this.refresh_ instanceof Deferred) {
				console.warn("login called while refreshing");
				return this.refresh_;
			}
		
			lastLogin = { user: user, pass: pass };
			
			var d = new Deferred();
	        ajax(this.host + this.baseUrl + "session/", {
	            method: "POST",
	            contentType: "application/json",
	            data: JSON.stringify({ username: user, password: pass })
	        }).then(function(resp) {
				resp = typeof resp === "string" ? JSON.parse(resp) : resp;

				delete Session.login_;

	        	if((lastLogin.success = resp.success) === false) {
	        		return d.callback(resp);
	        	}
	        	
				Session.hash = resp.id;
				Session.startedAt = new Date();
				Session.details = resp;
				
				lastLogin.success = true;
        		lastLogin.timeout = Date.now() + 5 * 60 * 1000;
				
				delete Session.refresh_;
				
				if(Session.waitFor_ instanceof Deferred) {
					if(Session.waitFor_.fired === -1) {
						Session.waitFor_.callback(Session.hash);
					}
				}

				d.callback(resp);
	        })
	        return (this.login_ = d);
		},
		logout: function() {
			var d = new Deferred(), hash = Session.hash;

			["waitFor_", "refresh_", "info_", "available_", "login_", "hash", "details"].forEach(function(k) {
				if(Session[k] instanceof Deferred) {
					try {
						// Session[k].addErrback(function(err) { console.log(">>>>", err); return 0; }).cancel();
					} catch(e) {
						// console.log(k, e.message, e);
						console.log(" >>>> " + k);
						return true;
					}
				}
				delete Session[k];
			});
			
			if(hash !== undefined) {
				ajax(this.host + this.baseUrl + "session/" + hash, { method:"DELETE" })
					.then(function() { d.callback(true); })
					.fail(function() { d.callback(true); });
			} else {
				d.callback(false);
			}
			return d;
		}
	});
});