define(function(require) {

	var VO = require("./VO");
	var EM = require("veldoffice/EM");

	var Apg = require("json!sikb/9.1.0/Analyseparameter-groep");
	var ApgMap = Apg.reduce((a, o) => { a[o.sikbId] = o; return a;}, {});

	var to_a_map = (acc, item) => { acc[item.ID] = item; return acc; };
	var Toetsoordelen = require("json!sikb/13.5/imsikb0101/Toetsoordelen").reduce(to_a_map, {});

	function resolveXlinkHrefs(feature, map) {
	
		function mapFeatures(feature, map, done) {
			done = done || [];
			if(done.indexOf(feature) === -1) {
				done.push(feature);
				for(var key in feature) {
					var value = feature[key];
					if(key === "gml:id") {
						map[value] = feature;
					} else {
						if(typeof value === "object" && value !== null) {
							mapFeatures(value, map);
						}
					}
				}
			}
			return map;
		}
		
		map = map || mapFeatures(feature, {});
	
		var value;
		for(var key in feature) {
			if(key === "xlink:href") {
				value = feature[key];
				if(value.charAt(0) === "#") value = value.substring(1);
				feature['xlink:href-resolved'] = map[value];
			} else {
				value = feature[key];
				if(typeof value === "object" && value !== null) {
					resolveXlinkHrefs(value, map);
				}
			}
		}
		
		return map;
	}
	function associate(instance, type, feature) {
		var arr;
		
		if(!(arr = instance[type])) {
			instance[type] = [feature];
		} else {
			if(!arr.includes(feature)) {
				var types = type.split(":");
				if(types[0] === "immetingen" || types[0] === "imsikb0101") {
					if(arr.filter(_ => _['gml:id'] === feature['gml:id']).length > 0) {
						return;
					}
				}
				arr.push(feature);
			}
		}
	}
	function findLovcomponentByCA(ca) {
		var xid = js.get(["immetingen:physicalProperty", "immetingen:PhysicalProperty", 
			"immetingen:parameter", "_Data"], ca);
			
		xid = parseInt(xid.split(":").pop(), 10);

		// var GB = "geobis:Analyseparameter.sikbId", map;
		// if(!(map = EM.instances.Lovcomponent._map)) {
		// 	map = (EM.instances.Lovcomponent._map = {});
		// 	Object.keys(EM.instances.Lovcomponent).map(_ => { 
		// 		map[_.xid || js.get(GB, _)] = _; 
		// 	});
		// }
		
		var r = EM.get("Lovcomponent", "_" + xid);
		if(r.xid === undefined) {
			r.xid = xid;
			r.naam = (r['geobis:Analyseparameter'] = ApgMap[xid] || {}).omschrijving || "?";
		}
		return r;
	}
	
	// cache in memory	
	var vats = {}, walks = {}, printer;
	
	return js.mixIn(VO, {
		vat: (id, refresh) => {
			if(!refresh && vats[id]) return vats[id];

			var walk = (id, paths) => VO.json(js.sf(
					"/office-rest/action/rapportage?view=Veldoffice-walk&entity=Onderzoek&id=%s&log", 
						id));
			var vat = (id) => VO.json(js.sf(
					"/office-rest/action/rapportage?id=%s&view=Onderzoek-VAT", id));

			return (vats[id] = Promise.all([vat(id), walk(id)]).then((all) => {
				if(typeof all[0] === "string") return all[0];//throw new Error(all[0]);
				if(typeof all[1] === "string") return all[1];//throw new Error(all[1]);
				
				var vat = all[0], walk = EM.processWalkResult2(all[1]);
				
				function resolve(key, arr, fn, idx) {
					return (arr || []).map(index => fn(vat[key].alle[index][idx]));
				}
				
				vat.bodemmonsters.map(bm => {
					js.mixIn(EM.get("Bodemmonster", bm[0]), {
						'.analysepakketten': resolve("pakketten", bm[5], (id) => EM.get("AnalysePakket", id), 1),
						'.analyseresultaten': resolve("resultaten", bm[6], (id) => EM.get("AnalyseResultaat", id), 2)
					});
				});
				vat.watermonsters.map(wm => {
					js.mixIn(EM.get("Watermonster", wm[0]), {
						'.analysepakketten': resolve("pakketten", wm[6], (id) => EM.get("AnalysePakket", id), 1),
						'.analyseresultaten': resolve("resultaten", wm[7], (id) => EM.get("AnalyseResultaat", id), 2)
					});
				});
				vat.toetsingen.filter(tn => tn.gml).map(tn => {
					var map = resolveXlinkHrefs(tn.gml);
					var log = (tn.log = []), ams = (tn.monsters = []), ars = (tn.resultaten = []);
					log.push = function() {
						var args = arguments.length !== 2 ? arguments : [
							Object.create(arguments[1],
								{ name: { value: arguments[0] }
							})
						];
						return Array.prototype.push.apply(this, args);
					};

					// `#VA-20200904-1`
					Array.as(js.get("immetingen:Sample", tn.gml)||[]).map((sample, si, sa) => {
						var gid = js.get("gml:id", sample); // if(!gid || vat.ids[gid]) debugger;
						var am = EM.get("Analysemonster", vat.ids[gid] || gid);
						
						if(!vat.ids[gid]) {
							console.log(gid);
						}
						
						var ar, ca;
						
						sample['imsikb0101:toetsing'] = tn;
						
						var ditishet;
	// TODO initialize am (if necessary)
						if((ditishet = (am.naam === undefined))) {
							ams.push(am);
							am.naam = js.get("immetingen:name", sample);
							am.labAnalysemonster = EM.get("LabAnalysemonster", am.id + "_LA");
							am.labAnalysemonster.analysemonster = am;
							log.push("new Analysemonster", am);
						}

						sample['va:Analysemonster'] = am;
						am['immetingen:Sample'] = sample;
						
	// process in specific order (1st CalculatedAnalysis and then TestingConclusion)
						for(var ii = 0; ii < 2; ++ii) {
							Array.as(js.get("sam:relatedObservation", sample)||[]).map((ro, i, arr) => {
								var type = Object.keys(ro)[0]; // probably CalculatedAnalysis or TestingConclusion
								var ft = ro[type]; 
								var om_ro = js.get("om:relatedObservation", ft);
								if(om_ro instanceof Array) {
									if(ii === 0 && type === "immetingen:CalculatedAnalysis") {
	// TODO maybe this case is not really necessary
										log.push(js.sf("[%d/%d|%s][%d/%d] multiple (%d) om:relatedObservation (som-parameter)", 
												si, sa.length, am.naam, i, arr.length, om_ro.length), 
												{am: am, arr: om_ro, type: type, ft: ft, om_ro: om_ro});
										
	// this result doesn't exist in Veldoffice - create an inline AnalyseResultaat		
										ar = EM.get("AnalyseResultaat", ft['gml:id']);
										ft['va:AnalyseResultaat'] = ar;
										ars.push(ar);
	// TODO initialize more of ar's attributes
										ar.meetwaarde = null; ar.referentie = null;
										ar.analysemonster = am.labAnalysemonster;
										am.labAnalysemonster.analyseresultaten = am.labAnalysemonster.analyseresultaten || [];
										am.labAnalysemonster.analyseresultaten.push(ar);
										ar.component = findLovcomponentByCA(ft)
										associate(ar, "immetingen:CalculatedAnalysis", ft);
	
										om_ro.map(_ => {
											gid = js.get("om:ObservationContext.om:relatedObservation.xlink:href", _);
											if(gid.charAt(0) === "#") gid = gid.substring(1);
											js.set("om:ObservationContext.om:relatedObservation.xlink:href-resolved/AnalyseResultaat--", EM.get("AnalyseResultaat", vat.ids[gid] || gid), _);
										});
										
										associate(am, type, ft);
									} else if(ii === 0) {
										log.push(js.sf("[%d/%d|%s][%d/%d] not handled (run %d)", 
												si, sa.length, am.naam, i, arr.length, ii + 1), 
												{ft: ft, type: type});
									}
								} else {
									if((gid = js.get("om:ObservationContext.om:relatedObservation.xlink:href", om_ro))) {
										gid = gid.substring(gid.charAt(0) === "#" ? 1 : 0);
										if(ii === 0 && type === "immetingen:CalculatedAnalysis") {
											if(ditishet) {
												ar = EM.get("AnalyseResultaat", gid);
												ft['va:AnalyseResultaat'] = ar;
												ars.push(ar); 
		// TODO initialize more of ar's attributes
												ar.meetwaarde = null; ar.referentie = null;
												ar.analysemonster = am.labAnalysemonster;
												am.labAnalysemonster.analyseresultaten = am.labAnalysemonster.analyseresultaten || [];
												am.labAnalysemonster.analyseresultaten.push(ar);
												ar.component = findLovcomponentByCA(ft)
												associate(ar, "immetingen:CalculatedAnalysis", ft);
												
												log.push(js.sf("[%d/%d|%s][%d/%d] %s associated with extra va:AnalyseResultaat - %s", 
													si, sa.length, am.naam, i, arr.length, type,
													(js.get("immetingen:physicalProperty.immetingen:PhysicalProperty.immetingen:parameter._Data", ft)||"?").split(":").pop()), 
													{ca: ft, ro: ro});
											} else {
												ar = EM.get("AnalyseResultaat", vat.ids[gid] || gid);
												ft['va:AnalyseResultaat'] = ar;
												associate(ar, type, ft);
												js.set("om:ObservationContext.om:relatedObservation.xlink:href-resolved/AnalyseResultaat!", ar, om_ro);
												
												log.push(js.sf("[%d/%d|%s][%d/%d] %s associated with va:AnalyseResultaat", 
														si, sa.length, am.naam, i, arr.length, type), {ar: ar, type: type, ft: ft});
												
											}
										} else if(ii === 1 && type === "imsikb0101:TestingConclusion") {
											// TestingConclusion-s refer to CalculatedAnalysis, find that ca by gid
											ca = arr.filter(_ => _[Object.keys(_)[0]]['gml:id'] === gid)[0];
											ca = ca['immetingen:CalculatedAnalysis'];
											ar = ca['va:AnalyseResultaat'];
											if(ar !== undefined) {
												associate(ar, type, ft);
												log.push(js.sf("[%d/%d|%s][%d/%d] %s associated with va:AnalyseResultaat", 
														si, sa.length, am.naam, i, arr.length, type), {ar: ar, type: type, ft: ft});
											} else {
												log.push(js.sf("[%d/%d|%s][%d/%d] va:AnalyseResultaat missing for %s", 
														si, sa.length, am.naam, i, arr.length, type), {ca: ca, ft: ro});
											}
										}
									} else if(ii === 1) {
										associate(am, type, ft); //ro
										log.push(js.sf("[%d/%d|%s][%d/%d] %s associated with va:Analysemonster", 
												si, sa.length, am.naam, i, arr.length, type), ro);
									}
								}
							});
						}
					});
				});
				
	// it has been very confusing several times now to not have the actual instances (while inspecting in the console)
				walk.instances_EM = {};
				for(var e in walk.instances) {
					for(var k in walk.instances[e]) {
						js.set(js.sf("%s.%s", e, k), EM.get(e, k), walk.instances_EM);
					}
				}
				walk.instances = walk.instances_EM;
				delete walk.instances_EM;
				
				EM.instances.Deelmonster = EM.instances.Deelmonster || {};
				EM.instances.Bodemmonster = EM.instances.Bodemmonster || {};
				EM.instances.Watermonster = EM.instances.Watermonster || {};
				EM.instances.AnalyseResultaat = EM.instances.AnalyseResultaat || {};

				Object.values(EM.instances.Deelmonster).forEach(dm => {
					var bm = EM.instances.Bodemmonster[dm.veldmonster.id];
					var wm = EM.instances.Watermonster[dm.veldmonster.id];

					// bm && associate(dm.veldmonster, "va:Bodemmonster", bm);
					// wm && associate(dm.veldmonster, "va:Watermonster", wm);
					
					if(bm) { dm.bodemmonster = bm; }
					if(wm) { dm.watermonster = wm; }
					
					bm && associate(bm, "va:Veldmonster", dm.veldmonster);
					wm && associate(wm, "va:Veldmonster", dm.veldmonster);
				});
				Object.values(EM.instances.AnalyseResultaat).reduce((a, o) => { a.push(o); return a; }, [])
					.filter(ar => ar['imsikb0101:TestingConclusion'])
					.forEach(ar => {
						ar['imsikb0101:TestingConclusion'].forEach(tc => {
							var to = js.get("om:result.imsikb0101:ClassifiedResult.imsikb0101:classifiedResult._Data", tc);
							to = to && to.split(":").pop();
							if((to = Toetsoordelen[to]) !== undefined) {
								var oordelen = js.get("analysemonster.analysemonster.imsikb0101:toetsoordelen", ar, {});
								var T = js.get("imsikb0101:TestingConclusion.0.imsikb0101:testing.imsikb0101:ExecutedTesting.imsikb0101:testingType._Data", ar) || js.get("imsikb0101:TestingConclusion.0.imsikb0101:testing.xlink:href-resolved.imsikb0101:testingType._Data", ar);
								var key = js.sf("T%d:%d", T.split(":").pop(), to.ID);
								var O = (oordelen[key] = oordelen[key] || { 
									'.': to, 
									'va:analyseresultaten': [],
									'imsikb0101:toetsing': T
								});
								var arr = O['va:analyseresultaten'];
								if(!arr.includes(ar)) arr.push(ar);
							}
						})
					});

				return {V: vat, W: walk};
			}));
		}
	});
});
