define(function(require) {

	var REPORTS = "/office-rest/action/rapportage";
	var EM = require("veldoffice/EM");
	var Session = require("veldoffice/Session");
	var Hash = require("util/Hash");

	var walks = EM.walks;

	function uuidv4() {
		return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, (c) => {
			var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
			return v.toString(16);
		});
	}
	
	var U = localStorage.vousr, P = localStorage.vopwd;
	var VO = {
		ROOT_WALK: REPORTS,
		em: EM,
		session: Session,  // symmetry!
		instances: EM.instances,
		entities: EM.entities,
		models: EM.models,
		walks: walks,
		// codes, coderingen?
		
		// session
		li: (u, p) => Session.login(u||U, p||P),
		lo: () => Session.logout(),
		
		// data
		json_: (url) => fetch(url).then(res => res.text()).then((text) => {
			try { return JSON.parse(text); } catch(e) {
				throw text; // why are we so forgiving? maybe better throw?
			}
		}),
		json: (url) => Session.get(url),
		walk: (entity, key, paths, opts, force, result) => {
			var k = entity + "-" + key, view = true;
			if(paths instanceof Array) {
				paths = paths.join(",");
				view = false;
			}
			
			k += js.sf("[%s]", paths || ".");

			if(!result) {
				if(force || !walks[k]) {
					if(view) {
						var url = js.sf("%s?view=%s&entity=%s&id=%s&log", VO.ROOT_WALK, paths, entity, key);
						walks[k] = VO.json(url).then(res => {
							EM.processWalkResult2(res);
							VO.resolve_fotos(entity, key, walks[k]);
							walks[k] = res;
							return res;
						}).catch(res => {
							delete walks[k];
							throw res;
						});
					} else {
						var url = js.sf("%s?view=Veldoffice-walk&entity=%s&id=%s&paths=%s&log", VO.ROOT_WALK, entity, key, paths || ".");
						walks[k] = VO.json(url).then(res => {
							EM.processWalkResult2(res);
							VO.resolve_fotos(entity, key, walks[k]);
							walks[k] = res;
							return res;
						}).catch(res => {
							delete walks[k];
							throw res;
						});
						
					}
				}
			} else {
				VO.resolve_fotos(entity, key, (walks[k] = result));
			}
			return Promise.resolve(walks[k]).then(_ => VO.instances[entity][key]); 
		},
		
		// util
		// TODO Toetsing: VO.vat() => some util stuff should refactor/go here?
		resolve_fotos: (entity, key, instance) => {
			if(instance.fotos instanceof Array) {
				instance.fotos.forEach(foto => {
					var arr, ins = js.get([foto.type.split(".").pop(), foto.entityId], VO.em.instances);
					if(ins) {
						if((arr = js.get("vo:fotos", ins, [])).indexOf(foto) === -1) {
							// console.log("linked foto", foto, "to", js.nameOf(ins), ins);
							arr.push(foto);
						}
					} else {

						for(var e in VO.em.instances) {
							if((ins = VO.em.instances[e][foto.entityId])) {
								if((arr = js.get("vo:fotos", ins, [])).indexOf(foto) === -1) {
									// console.log("linked thoroughly?");
									arr.push(foto);
								}
								return;
							}
						} 
						// console.warn("could not link foto", foto.entityId, foto.onderzoek.id, foto.type, foto);
					}
				});
			}
			return instance;
		},
		
		json2document: (onderzoek, content, opts) => {
			if(typeof content !== "string") {
				content = JSON.stringify(content);
			}
			var replacements = {
				"ö": "&#214;", "ë": "&#235;",
				"σ": "&#x03C3;",//"&#963;",
				"ε": "&#949;",
				"ρ": "&#x03C1;", // rho
				"→": "&#8594;", "↑":"&#8593;",
				"µ": "&#181;", "√": "&#8730;",
				"α": "&#945;"
			};
			for(var k in replacements) {
				content = content.replace(new RegExp(k, "g"), replacements[k]);
			}

			var naam = opts.naam || Hash.md5(content) + ".json";
			
			// checks whether document already exists
			return VO.em.query("Document:onderzoek/" + onderzoek.id + "/document", { 
				where: ["equals", "naam", naam] 
			}).then(res => {
				if(res.length === 0) {
					var formData = new FormData();
					var blob = new Blob([content], { type: "text/xml"});
					formData.append("file", blob, naam);
				
					opts = opts || { omschrijving: "Temporary-" + Date.now() };
				
					formData.append("onderzoekId", onderzoek.id);
					formData.append("omschrijving", js.sf("%s {size:%d}", opts.omschrijving, content.length));
					formData.append("autoRemove", opts.hasOwnProperty("autoRemove") ? opts.autoRemove : true);
				
					return new Promise((resolve, reject) => {
						var request = new XMLHttpRequest();
						request.open("POST", "/office-rest/action/documenten/upload");
						request.send(formData);
						request.addEventListener("readystatechange", (e) => {
							// console.log(js.sf("%s - %s", e.target.status, e.target.statusText), e);
							if(e.target.status === 200) {
								resolve(naam);
							} else {
								reject(e);
							}
						});
					});
				}
				
				return res[0].naam;
			});
			
		}
	};
	
	function opts2params(opts) {
		if(!!opts === false) {
			return "";
		}

		if(typeof opts === "boolean") return "&test=true";
		
		if(typeof opts === "object") {
			return "&" + Object.keys(opts).map(function(key) {
				return js.sf("%s=%s", window.escape(key), window.escape(opts[key]));
			}).join("&");
		}
		
		return opts;
	}
	
	VO.q = (e, a, c) => VO.em.query(e, a, c).then(res => [res, 
		H("devtools/Alphaview", { 
			sel: [res], 
			label: "VO.q('" + e + "')" 
		})
	]);
	VO.itsme = VO.li;
	VO.exec = VO.view = (onderzoek, view, opts) => VO.json(js.sf("%s?id=%d&view=%s%s", 
		VO.em.ROOT_WALK, onderzoek, view, opts2params(opts)));
	VO.checkReload = (re, page) => fetch((page || "index.html") + "?" + Date.now()).then(res => res.text().then(text => {
		var lhs = text.match(re || /urlArgs\:\s"(.*)"/);
		if(lhs && lhs.length === 2) {
			var key = window.location.pathname + "-reloading";
			var rhs = window.require.s.contexts._.config.urlArgs;
			lhs = lhs.pop();
			if(typeof rhs === "function") {
				rhs = rhs("","").substring(1);
			}
			if(lhs !== rhs) {
				console.log("New version detected", lhs);
				var last = localStorage.getItem(key);
				if(last === null) {
					localStorage.setItem(key, Date.now());
					window.setTimeout(() => window.location.reload(), 0);
				} else {
					console.log("Reloading application blocked. Reset after 10s.");
					if(Date.now() - parseInt(last, 10) > 1000*10) {
						localStorage.removeItem(key);
						window.setTimeout(() => window.location.reload(), 0);
					}
				}
			} else {
				localStorage.removeItem(key);
			}
		} else {
			console.warn("Version of application could not be detected.");
		}
	}));
	VO.util = {
		Foto_hasRapportage: (foto) => (foto.rapportage !== "false") && (foto.rapportage !== false),
		Meetpunt_sort: (m1, m2) => (m1 = js.get("code", m1)) < (m2 = js.get("code", m2)) ? -1 : (m1 > m2 ? 1 : 0),
		
		Credentials_reload: () => { U = localStorage.vousr; P = localStorage.vopwd; delete Session.login_; },
		
		UI_fadeOut: (control, callback) => {
			control.addClass("fadeout");
			control.set("css", {
				".fadeout": "transition: opacity 1450ms ease-out;",
				".fadeout-end": "opacity:0;"
			});
			control.addClass("fadeout-end");
			control.setTimeout("fadeout", callback, 1750);
		},
		UI_collect: (control) => {
			var parent = control.getParent();
			parent.setTimeout("UI_collect", () => {
				if(!parent._controls || parent._controls.length === 0) {
					parent.destroy();
				}
			}, 2000);
		}
	};

	return VO;
});