define([
	"veldoffice/VO", "devtools/Resources", "veldapps-xml/index"
], (VO, Resources, Xml) => {
	
	const cache = {
		sikb135: {},
		sikb146: {},
		sikb148: {},
		sikb149: {},
		document: {}
	};
	const get_ = Resources.get, link_ = Resources.link;
	
	const get_codes = () => {
		var url = (uri) => "/office-rest/rest/v1/" + uri;
		var str = (s) => new window.String(s); // CVLN-20210113-1
		var result = {};
		
		return fetch(url("codering"))
			.then(res => res.json())
			.then(res => res.filter(_ => _.name !== "Codering").sort((i1, i2) => i1.name < i2.name ? -1 : 1))
			.then(res => Promise.all(res.map(codering => {
				var list = fetch(url("codering/" + codering.name)).then(res => res.json());
				// this.print(str(codering.name), list);
				return list.then(res => {
					result[codering.name] = res;
					return res;
				});
			})))
			.then(res => {
	    		const text = JSON.stringify(result);
	    		return {
		    		name: "Veldoffice-codes",
		    		uri: "veldoffice://codes",
		    		size: text.length,
		    		text: text,
		    		type: "File",
		    		contentType: "application/json",
		    		generated: Date.now()
		    	};
	    	});
	};

	Resources.get = function(uri) {
	    var id;
	
	    function handleSIKBCase(cacheKey, version) {
	        const path = uri.split("/");
	        id = path[4];
	
	        if (cache[cacheKey][id]) return cache[cacheKey][id];
	
	        var ozk = VO.em.instance("Onderzoek", id);
	        var onderzoek = [ozk.projectcode, ozk.naam].filter(v => v).join(" - ");
	
	        return (cache[cacheKey][id] = fetch(js.sf("/office-rest/action/rapportage/?id=%s&export=SIKB&version=" + version, id))
	            .then(resp => resp.text().then(text => ({
	                name: onderzoek ? js.sf("%s - %s", onderzoek, `SIKB-${version}-export.xml`) : path[5],
	                uri: onderzoek ? js.sf("%s/%s", uri, onderzoek) : uri,
	                size: text.length,
	                text: text,
	                type: "File"
	            }))))
	            .catch(e => { delete cache[cacheKey][id]; return e; });
	    }
	
	    if (uri.startsWith("veldoffice://onderzoek/sikb13/") || uri.startsWith("veldoffice://onderzoek/sikb135/")) {
	        return handleSIKBCase('sikb135', '13.5.0');
	    } else if (uri.startsWith("veldoffice://onderzoek/sikb146/")) {
	        return handleSIKBCase('sikb146', '14.6.0');
	    } else if (uri.startsWith("veldoffice://onderzoek/sikb148/")) {
	        return handleSIKBCase('sikb148', '14.8.0');
	    } else if (uri.startsWith("veldoffice://onderzoek/sikb149/")) {
	        return handleSIKBCase('sikb149', '14.9.0');
	    } else if (uri.startsWith("veldoffice://onderzoek/document/")) {
	        id = uri.split("/")[4];
	
	        if (cache.document[id]) return cache.document[id];
	
	        var doc = VO.em.instance("Document", id);
	
	        return (cache.document[id] = fetch(js.sf("/office-rest/action/documenten/view?id=%s", id))
	            .then(resp => resp.text().then(text => ({
	                name: doc.naam,
	                uri: doc.naam ? uri + "/" + doc.naam : uri,
	                size: text.length,
	                text: text,
	                type: "File"
	            }))))
	            .catch(e => { delete cache.document[id]; return e; });
	    } else if(uri.startsWith("veldoffice://codes")) {
	    	return get_codes();
	    }
	
	    return get_.apply(this, arguments);
	};
	Resources.link = function(uri) {
		if(uri.startsWith("veldoffice://onderzoek/document/")) {
			var id = uri.split("/")[4];
			var doc = VO.em.instance("Document", id);
			return Promise.resolve(js.sf("/office-rest/action/documenten/view?id=%s", id));
		}
		return link_.apply(this, arguments);
	};
	
	// done and done - Veldoffice documents integrated by extension of "filename"  :-D

	VO.q = (e, a, c) => VO.em.query(e, a, c).then(res => [res, H("devtools/Alphaview", { sel: [res] })]);
	VO.resource2doc = (uri) => {	
		const isfolder = uri.endsWith("/");
		if(isfolder) uri = uri.substring(0, uri.length - 1);

		const doc = VO.em.get("Document", uri);
		doc.naam = uri;
		doc.resource_ = { 
			name: uri,
			uri: uri, 
			type: isfolder ? "Folder" : "File"
		};
		
		return doc;
	};

	return {
		init: () => Promise.resolve(true),
		
		VO: VO,
		Resources: Resources,
		Xml: Xml
	};
});

/*- 

The Resources object in the code is being used to load resources such as images, fonts, and other files required by the application. The code overrides the get and link methods of the Resources object to intercept calls to resources that have a specific URI format starting with "veldoffice://onderzoek/document/".

When such a URI is encountered, the overridden get and link methods handle the resource loading differently. They parse the ID of the document from the URI, use the ID to fetch the corresponding document data from a REST endpoint, and return the document data as if it were a regular resource.

This way, the code is able to integrate Veldoffice documents into the application by extending the "filename" of the document, without having to modify the application's existing resource loading logic.

*/