define(function(require) {

	require("leaflet");
	require("leaflet/plugins/proj4");

	var Layers = require("leaflet/layers/28992");
	var Browser = require("util/Browser");
	var HE = require("util/HtmlElement");
	var Crs = require("./Crs");
	
	return {
		initialize: function(mapDomNode, mapOptions, options) {
			var touch = "ontouchstart" in window;
			
			mapOptions = js.mixIn({ 
				zoomControl: false, 
				attributionControl: false,
				preferCanvas: !true 
			}, mapOptions || {});
            
            if(mapOptions.crs === undefined) {
            	// Assume RD by default (for Veldoffice anyways)
            	mapOptions.crs = Crs.RD;
            } else if(typeof mapOptions.crs === "number") {
            	mapOptions.crs = Crs[mapOptions.crs];
            }
			
            var map = L.map(mapDomNode, mapOptions);
            options = options || {};

		    if(Browser.ios !== true && Browser.android !== true) {
				/*- TODO move to Map.page */
				HE.addClass(mapDomNode, "desktop");
		    }

			if(options.layers === true) {		    
				V7.layerControl = L.control.layers({
					"BRT Achtergrondkaart": Layers.brtachtergrondkaart(),
					"Open Basis Kaart": Layers.openbasiskaart().addTo(map) 
			}, {
					"Kadastrale kaart": Layers.kadastralekaartv2(),
					"BGT Lijngericht": Layers.bgtlijngericht(),//.addTo(map),
					"BGT Pastel": Layers.bgtpastel().addTo(map)
				}).addTo(map);
			} else {
				Layers.openbasiskaart().addTo(map);
				// L.control.layers(options.layers || {
				// 	openbasiskaart: Layers.openbasiskaart().addTo(map),
				// 	// luchtfoto2017: Layers.luchtfoto2017().addTo(map),
				// 	// bgtlijngericht: Layers.bgtlijngericht().addTo(map),
				// 	// bgtpastel: Layers.bgtpastel().addTo(map)
				// }).addTo(map);
			}
			
			if(options.scale !== false) {
	            L.control.scale({imperial: false}).addTo(map);
			}

			if(typeof V7 !== "undefined") {
				// FIXME remeove V7 dependencies
				V7.locationControl = L.control.locate({
					follow: true, 
					stopFollowingOnDrag: true
				}).addTo(map);
			}
			
			return map;
		}
	};

});