define(function(require) {
	
	var Base = require("vcl/entities/Query");
	var Query = require("js/defineClass");
	var Type = require("js/Type");
	var EM = require("veldoffice/EM");
	var Source = require("data/Source");
	var SourceEvent = require("data/SourceEvent");
	
	function put(keys, db, instance, id, doc) {
		doc = doc || {_id: id};
		keys.forEach(function(k) {
			if(instance.hasOwnProperty(k + "$")) {
				doc[k] = instance[k + "$"];
			} else {
				doc[k] = instance[k];
			}
		});
		
		// db.put(updated version);
		delete doc.id;
		return db.put(doc);
	}
	
	return (Query = Query(require, {
		inherits: Base,
		prototype: {
			// _count: false,
			_size: false, // lame way to block request, until refresh(size) has been called
			_path: "",
			_raw: false,
			_options: null,
			
			_cache: false,
			_instances: null,
			
			_summary: null,
			
			_limit: 50,

			getCacheDb: function() {
				var PouchDB = require("pouchdb");
				if(this._cache !== false && !this._cache.db) {
					this._cache.db = new PouchDB(this._cache.name);
				}
				return this._cache.db;
			},
			readFromCache: function() {
				var db = this.getCacheDb();
				db.allDocs().then(function() {
					
				});
			},
			cacheResult: function(res, page) { 
				var me = this, db = this.getCacheDb(), keys = this.getAttributeNames();
				
				var instances = [].concat(res.instances);
				if(this._instances) {
					this._instances.push.apply(this._instances, instances);
				} else {
					this._instances = instances;
					this._instances.start = Date.now();
					this._instances.inserted = 0;
					next();
				}
				
				function next() {
					var instances = me._instances;
					if(instances && instances.length) {
						var instance = instances.shift();	
						var id = "" + instance.id;
						db.get(id).then(function(doc) {
							put(keys, db, instance, id, doc).then(next).catch(
								function() { 
									console.log("catch", arguments); 
									next(); 
								});
						}).catch(function() { 
							put(keys, db, instance, id).then(next).catch(
								function() { 
									console.log("catch!", arguments); 
									next(); 
								});
						});
						
						instances.inserted++;
						
						if(instances.length % 100 === 0) {
							console.log(me._name, 
								instances.inserted + " done", 
								instances.length + " to go", 
								Date.now() - me._instances.start);
						}
						
					} else {
						console.log(me._name, "done caching", Date.now() - me._instances.start);
						delete me._instances;
					}
				}


/*
				var start = Date.now();
				res.instances.forEach(function(instance, i) {
					var id = "" + instance.id;
					
					function done() {
						if(i === res.instances.length - 1) {
							me._inserted += res.instances.length;
							console.log(this._name, "inserted", me._inserted / me._queued, Date.now() - start);
						}
					}					
					
					db.get(id).then(function(doc) {
						put(keys, db, instance, id, doc).then(done);
					}).catch(function() { 
						put(keys, db, instance, id).then(done);
					});
				});
*/
			},
			isValid: function() {
				if(this.isLoading()) { return false; }
				
				if(this._entity !== "" || this._path !== "") {
					if(this._layoutChanged === true) {
						delete this._layoutChanged;
						this.notify(SourceEvent.layoutChanged);
					}
				}

				return this._path !== "" || (this._entity !== "" || this._raw === true);
			},
			getSize: function() {
				/** @overrides */
				this.requestWhenNeeded();
				
				var size;
				if(this._size === false) {
					size = (this._arr && this._arr.length) || 0;
				} else {
					size = this.inherited(arguments);
					// if(size < this._size) { why were we doing this? breaks onFilterObject
					// 	size = this._size;
					// }
				}
				return size;
			},
			getEM: function() {
				return EM;
			},
			getRequestCriteria: function(page) {
				// if entity is not set, default to raw=true
				var raw = this._raw || (this._entity === "");
				var criteria = {
					// only the first request actually does a count request
					count: this._arr === null && (typeof this._count === "object" ? js.mixIn(this._count) : this._count),
					// shouldn't a 'page' be identified like: [page, size]; 'en-una-vez'?
					page: page, pagesize: this._limit,
				};
				
				criteria.opts = js.mixIn(this._options || {});

				if(typeof criteria.count.where === "function") {
					criteria.count.where = criteria.count.where.apply(this, [page]);
				}
				
				if(this._path || raw) {
					if(this._path) { criteria.opts.path = this._path; }
					if(raw) { criteria.opts.raw = raw; }
				}
				if(this._where !== null) {
					if(typeof this._where === "function") {
						criteria.where = this._where.apply(this, [page]);
					} else {
						criteria.where = this._where;
					}
				}
				if(this._groupBy !== "") {
					criteria.groupBy = this._groupBy;//.split(",");
				} else if(this._groupBy !== false) {
					//automatically generate groupBy?
				}
				if(this._having !== null) {
					// TODO how does this feature work exactly?
					criteria.having = this._having;
				}
				if(this._orderBy !== "") {
					criteria.orderBy = this._orderBy;//.split(",");
				}
				if(this.hasOwnProperty("_onGetRequestCriteria")) {
					criteria = this.fire("onGetRequestCriteria", [criteria, page]);
				}
				return criteria;
			},
			processResult: function(res, page, criteria) {
				if(criteria.count) {
					var size = (this._size = res[0]);
					res = {
						count: res[0],
						summary: criteria.count.result, // 20211104 liking the term `summary`
						instances: res[1],
						tuples: []//res[1]
					};
/*
					this._tuples = [];
					while(size--) {
						this._tuples.push(Source.Pending);
					}
					this._arr = [].concat(this._tuples);
					this._array = [];
*/
				} else {
					res = {
						instances: res,
						tuples: []
					};
				}
				
				if(!(this._summary = res.summary)) {
					delete this._summary;
				}
				
/*- TODO cut of tuples the last page according to the size */
				if(this._size !== false && (page + 1) * this._limit > this._size) {
					// console.log((page + 1) * this._limit, ">", this._size);
					res.instances = res.instances.slice(0, this._size % this._limit);
				} else {
					// console.log((page + 1) * this._limit, "<", this._size);
				}
				
				var args = [res, page, criteria]; args.callee = arguments.callee;
				
				if(this._cache !== false) {
					this.cacheResult(res, page);
				}
				return this.inherited(args);
			},
			getRaw: function() {
				return this._raw;
			},
			setRaw: function(value) {
				if(this._raw !== value) {
					this._raw = value;
					if(this.isLoading()) {
						//this._changed = true;
					} else {
						this.notify(SourceEvent.changed);
					}
				}
			},
			getPath: function() {
				return this._path;
			},
			setPath: function(value) {
				if(this._path !== value) {
					this._path = value;
					if(this.isLoading()) {
						//this._changed = true;
					} else {
						this.notify(SourceEvent.changed);
					}
				}
			},
			refresh: function(size) {
				this._size = size;
				return this.inherited(arguments);
			},
			fetchAll: function(size, maxPages) {
				/*- calling refresh with a size, will initialize this._arr, 
					and moreover it will not request anything */
				this.refresh(size);
				/* TODO What about not downloading? */
				
				var pages = Math.round(size / this._limit);
				if(pages > maxPages) pages = maxPages;
				for(var i = 0; i <= pages; ++i) {
					this.requestPage(i);
				}
			},
			getAttributeNames: function() {
				/** @overides */
				return this.inherited(arguments).map(function(s) {
					return s.split(" ").pop();
				});
			},
			getAttributeValue: function(name, index) {
				/** @overrides ../data/Source.prototype.getAttributeValue */
				// return this.inherited(arguments, 1);
				
				// TODO based upon inherited method, is NOT calling inherited
				
				this.assertArray(index);
				if(name === ".") {
					return this.getObject(index || 0);
				}
				// var value = js.get(name, this.getObject(index || 0));
				var value = (this.getObject(index || 0) || {});
				
				if(value && value._prototype && value._prototype.hasOwnProperty(name)) {
					value = value._prototype[name];
				} else {
					value = js.get(name, value);
				}
				
				if(this._onGetAttributeValue !== null) {
					value = this.fire("onGetAttributeValue", [name, index, value]);
				}
				return value;
			},
			setAttributeValue: function() {
				return this.inherited(arguments, 1);
			},
			getOptions: function() { return this._options; },
			getSummary: function() {
				return this._summary;
			}
		},
		properties: {
			path: { type: Type.STRING, get: Function, set: Function },
			raw: { type: Type.BOOLEAN, get: Function, set: Function },
			options: { type: Type.OBJECT },
			cache: { type: Type.OBJECT }
		}
	}));
	
});